# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from distutils.version import StrictVersion

from bagpipe.bgp.common import logDecorator

from bagpipe.bgp.common.run_command import runCommand

from bagpipe.bgp.common.looking_glass import LookingGlassLocalLogger

from bagpipe.bgp.vpn.evpn import VPNInstanceDataplane
from bagpipe.bgp.vpn.dataplane_drivers import DataplaneDriver

from bagpipe.exabgp.message.update.attribute.communities import Encapsulation

BRIDGE_NAME_PREFIX = "evpn---"
VXLAN_INTERFACE_PREFIX = "vxlan--"
LINUX_DEV_LEN = 14


class LinuxVXLANEVIDataplane(VPNInstanceDataplane):

    def __init__(self, *args, **kwargs):
        VPNInstanceDataplane.__init__(self, *args)

        if 'linuxbr' in kwargs:
            self.bridge_name = kwargs.get('linuxbr')
        else:
            self.bridge_name = (
                BRIDGE_NAME_PREFIX + self.externalInstanceId)[:LINUX_DEV_LEN]

        self.vxlan_if_name = (
            VXLAN_INTERFACE_PREFIX + self.externalInstanceId)[:LINUX_DEV_LEN]

        self.log.info("EVI %d: Initializing bridge %s",
                      self.instanceId, self.bridge_name)
        if not self._interface_exists(self.bridge_name):
            self.log.debug("Starting bridge %s", self.bridge_name)

            # Create bridge
            self._runCommand("brctl addbr %s" % self.bridge_name)
            self._runCommand("brctl setfd %s 0" % self.bridge_name)
            self._runCommand("brctl stp %s off" % self.bridge_name)
            self._runCommand("ip link set %s up" % self.bridge_name)

            self.log.debug("Bridge %s created", self.bridge_name)

        self._create_and_plug_vxlan_if()

        self.log.debug("VXLAN interface %s plugged on bridge %s",
                       self.vxlan_if_name, self.bridge_name)

        self._cleaningUp = False

    @logDecorator.logInfo
    def cleanup(self):
        self.log.info("Cleaning EVI bridge and VXLAN interface %s",
                      self.bridge_name)

        self._cleaningUp = True

        self._cleanup_vxlan_if()

        # Delete only EVPN Bridge (Created by dataplane driver)
        if BRIDGE_NAME_PREFIX in self.bridge_name:
            self._runCommand("ip link set %s down" %
                             self.bridge_name, raiseExceptionOnError=False)
            self._runCommand("brctl delbr %s" %
                             self.bridge_name, raiseExceptionOnError=False)

    def _create_and_plug_vxlan_if(self):
        self.log.debug("Creating and plugging VXLAN interface %s",
                       self.vxlan_if_name)

        if self._interface_exists(self.vxlan_if_name):
            self._remove_vxlan_if()

        dstPortSpec = ""
        if self.driver.vxlanDestPort:
            dstPortSpec = "dstport %d" % self.driver.vxlanDestPort

        # Create VXLAN interface
        self._runCommand(
            "ip link add %s type vxlan id %d local %s nolearning proxy %s" %
            (self.vxlan_if_name, self.instanceLabel,
             self.driver.getLocalAddress(), dstPortSpec)
        )

        self._runCommand("ip link set %s up" % self.vxlan_if_name)

        # Plug VXLAN interface into bridge
        self._runCommand("brctl addif %s %s" % (self.bridge_name,
                                                self.vxlan_if_name))

    def _cleanup_vxlan_if(self):
        if self._is_vxlan_if_on_bridge():
            # Unplug VXLAN interface from Linux bridge
            self._unplug_from_bridge(self.vxlan_if_name)

        self._remove_vxlan_if()

    def _remove_vxlan_if(self):
        # Remove VXLAN interface
        self._runCommand("ip link set %s down" % self.vxlan_if_name)
        self._runCommand("ip link del %s" % self.vxlan_if_name)

    def _is_vxlan_if_on_bridge(self):
        (output, _) = self._runCommand(
            "brctl show %s | grep '%s' | sed -e 's/\s\+//g'" %
            (self.bridge_name, VXLAN_INTERFACE_PREFIX))

        return True if (output == self.vxlan_if_name) else False

    def _interface_exists(self, interface):
        """Check if interface exists."""
        (_, exitCode) = self._runCommand("ip link show dev %s" % interface,
                                         raiseExceptionOnError=False,
                                         acceptableReturnCodes=[-1])
        return (exitCode == 0)

    def _unplug_from_bridge(self, interface):
        if self._interface_exists(self.bridge_name):
            self._runCommand("brctl delif %s %s" %
                             (self.bridge_name, interface),
                             acceptableReturnCodes=[0, 1])

    def setGatewayPort(self, linuxif):
        gw_ip = self.gatewayIP
        gw_mac = "01:00:00:00:00:00"  # FIXME

        self._runCommand("brctl addif %s %s" %
                         (self.bridge_name, linuxif),
                         raiseExceptionOnError=False)

        self._runCommand("bridge fdb replace %s dev %s" %
                         (gw_mac, linuxif))

        self._runCommand(
            "ip neighbor replace %s lladdr %s dev %s nud permanent" %
            (gw_ip, gw_mac, linuxif)
        )

    def gatewayPortDown(self, linuxif):
        self._runCommand("brctl delif %s %s" %
                         (self.bridge_name, linuxif),
                         raiseExceptionOnError=False)
        # TODO: need to cleanup bridge fdb and ip neigh ?

    def setBridgeName(self, linuxbr):
        self.bridge_name = linuxbr

    @logDecorator.logInfo
    def vifPlugged(self, macAddress, ipAddress, localPort, label):
        # Plug localPort only if bridge was created by us
        if BRIDGE_NAME_PREFIX in self.bridge_name:
            self.log.debug("Plugging localPort %s into EVPN bridge %s",
                           localPort['linuxif'], self.bridge_name)
            self._runCommand("brctl addif %s %s" %
                             (self.bridge_name, localPort['linuxif']),
                             raiseExceptionOnError=False)

        self._runCommand("bridge fdb replace %s dev %s" %
                         (macAddress, localPort['linuxif']))

    @logDecorator.logInfo
    def vifUnplugged(self, macAddress, ipAddress, localPort, label,
                     lastEndpoint=True):
        # unplug localPort only if bridge was created by us
        if BRIDGE_NAME_PREFIX in self.bridge_name:
            self.log.debug("Unplugging localPort %s from EVPN bridge %s",
                           localPort['linuxif'], self.bridge_name)
            self._unplug_from_bridge(localPort['linuxif'])

    @logDecorator.log
    def setupDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri,
                                        encaps):
        if self._cleaningUp:
            self.log.debug("setupDataplaneForRemoteEndpoint: instance cleaning"
                           " up, do nothing")
            return

        mac = prefix
        ip = nlri.ip
        vni = label

        # populate bridge forwarding db
        self._runCommand("bridge fdb replace %s dev %s dst %s vni %s" %
                         (mac, self.vxlan_if_name, remotePE, vni))

        # populate ARP cache
        if ip is not None:
            self._runCommand("ip neighbor replace %s lladdr %s dev %s nud "
                             "permanent" % (ip, mac, self.vxlan_if_name))
        else:
            self.log.warning("No IP in E-VPN route, ARP will not work for this"
                             "IP/MAC")

        self._fdbDump()

    @logDecorator.log
    def removeDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri):

        if self._cleaningUp:
            self.log.debug("setupDataplaneForRemoteEndpoint: instance cleaning"
                           " up, do nothing")
            return

        mac = prefix
        ip = nlri.ip
        vni = label

        self._fdbDump()

        self._runCommand("ip neighbor del %s lladdr %s dev %s nud permanent" %
                         (ip, mac, self.vxlan_if_name))
        self._runCommand("bridge fdb del %s dev %s dst %s vni %s" %
                         (mac, self.vxlan_if_name, remotePE, vni))

        self._fdbDump()

    @logDecorator.log
    def addDataplaneForBroadcastEndpoint(self, remotePE, label, nlri, encaps):
        if self._cleaningUp:
            self.log.debug("setupDataplaneForRemoteEndpoint: instance cleaning"
                           " up, do nothing")
            return

        vni = label

        # 00:00:00:00:00 usable as default since kernel commit
        # 58e4c767046a35f11a55af6ce946054ddf4a8580 (2013-06-25)
        self._runCommand("bridge fdb append 00:00:00:00:00:00 dev %s dst %s "
                         "vni %s" % (self.vxlan_if_name, remotePE, vni))

        self._fdbDump()

    @logDecorator.log
    def removeDataplaneForBroadcastEndpoint(self, remotePE, label, nlri):
        if self._cleaningUp:
            self.log.debug("setupDataplaneForRemoteEndpoint: instance cleaning"
                           " up, do nothing")
            return

        vni = label

        self._fdbDump()

        self._runCommand("bridge fdb delete 00:00:00:00:00:00 dev %s dst %s "
                         "vni %s" % (self.vxlan_if_name, remotePE, vni))

        self._fdbDump()

    def _fdbDump(self):
        if self.log.debug:
            self.log.debug("bridge fdb dump: %s", self._runCommand(
                "bridge fdb show dev %s" % self.vxlan_if_name)[0])

    # Looking glass ####

    def getLookingGlassLocalInfo(self, pathPrefix):
        return {
            "linux_bridge": self.bridge_name,
            "vxlan_if": self.vxlan_if_name
        }


class LinuxVXLANDataplaneDriver(DataplaneDriver):

    """
    E-VPN Dataplane driver relying on the Linux kernel linuxbridge
    VXLAN implementation.
    """

    dataplaneInstanceClass = LinuxVXLANEVIDataplane
    requiredKernel = "3.11.0"
    encaps = [Encapsulation(Encapsulation.VXLAN)]

    def __init__(self, config, init=True):
        LookingGlassLocalLogger.__init__(self, __name__)

        self.log.info("Initializing %s", self.__class__.__name__)

        try:
            self.vxlanDestPort = int(config.get("vxlan_dst_port", 0)) or None
        except ValueError:
            raise Exception("Could not parse specified vxlan_dst_port: %s" %
                            config["vxlan_dst_port"])

        DataplaneDriver.__init__(self, config, init)

    def _initReal(self, config):
        self.config = config
        self.log.info("Really initializing %s", self.__class__.__name__)

        o = self._runCommand("uname -r")
        kernelRelease = o[0][0].split("-")[0]

        if (StrictVersion(kernelRelease) <
                StrictVersion(LinuxVXLANDataplaneDriver.requiredKernel)):
            self.log.warning("%s requires at least Linux kernel %s (you are"
                             " running %s)" %
                             (self.__class__.__name__,
                              LinuxVXLANDataplaneDriver.requiredKernel,
                              kernelRelease))

        self._runCommand("modprobe vxlan")

    def resetState(self):
        self.log.debug("Resetting %s dataplane", self.__class__.__name__)

        # delete all EVPN bridges
        cmd = "brctl show | tail -n +2 | awk '{print $1}'| grep '%s'"
        for bridge in self._runCommand(cmd % BRIDGE_NAME_PREFIX,
                                       raiseExceptionOnError=False,
                                       acceptableReturnCodes=[0, 1])[0]:
            self._runCommand("ip link set %s down" % bridge)
            self._runCommand("brctl delbr %s" % bridge)

        # delete all VXLAN interfaces
        cmd = "ip link show | awk '{print $2}' | tr -d ':' | grep '%s'"
        for interface in self._runCommand(cmd % VXLAN_INTERFACE_PREFIX,
                                          raiseExceptionOnError=False,
                                          acceptableReturnCodes=[0, 1])[0]:
            self._runCommand("ip link set %s down" % interface)
            self._runCommand("ip link delete %s" % interface)

    def _cleanupReal(self):
        # FIXME: need to refine what would be different
        self.resetState()

    def _runCommand(self, command, *args, **kwargs):
        return runCommand(self.log, command, *args, **kwargs)



# encoding: utf-8
"""
flow.py

Created by Thomas Mangin on 2010-01-14.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

from struct import pack

from bagpipe.exabgp.structure.address import Address,AFI,SAFI
from bagpipe.exabgp.structure.ip import Prefix
from bagpipe.exabgp.message.update.attributes import Attributes
from bagpipe.exabgp.message.update.attribute.id import AttributeID
from bagpipe.exabgp.message.update.attribute.communities import ECommunities

from bagpipe.exabgp.log import Logger
logger = Logger()

# =================================================================== Flow Components

class IComponent (object):
	# all have ID
	# should have an interface for serialisation and put it here
	pass

class CommonOperator:
	# power (2,x) is the same as 1 << x which is what the RFC say the len is
	power = { 0:1, 1:2, 2:4, 3:8, }
	rewop = { 1:0, 2:1, 4:2, 8:3, }
	len_position = 0x30

	EOL       = 0x80
	AND       = 0x40
	NOP       = 0x00

class NumericOperator (CommonOperator):
#	reserved  = 0x08
	LT        = 0x04
	GT        = 0x02
	EQ        = 0x01

class BinaryOperator (CommonOperator):
#	reserved  = 0x0C
	NOT       = 0x02
	MATCH     = 0x01

def _len_to_bit (value):
	return NumericOperator.rewop[value] << 4

def _bit_to_len (value):
	return NumericOperator.power[(value & CommonOperator.len_position) >> 4]

# Interface ..................

class IPrefix (IComponent):
	# not used, just present for simplying the nlri generation
	operations = 0x0
	ID = None
	NAME = None

	def __init__ (self,ipv4,netmask):
		self.nlri = Prefix(AFI.ipv4,ipv4,netmask)

	def pack (self):
		raw = self.nlri.pack()
		return "%s%s" % (chr(self.ID),raw)

	def __str__ (self):
		return str(self.nlri)

	def __repr__ (self):
		return str(self)

class IOperation (IComponent):
	# need to implement encode which encode the value of the operator

	def __init__ (self,operations,value):
		self.operations = operations
		self.value = value
		self.first = True

	def pack (self):
		l,v = self.encode(self.value)
		op = self.operations | _len_to_bit(l)
		if self.first:
			return "%s%s%s" % (chr(self.ID),chr(op),v)
		return "%s%s" % (chr(op),v)

	def encode (self,value):
		raise NotImplemented('this method must be implemented by subclasses')

#class IOperationIPv4 (IOperation):
#	def encode (self,value):
#		return 4, socket.pton(socket.AF_INET,value)

class IOperationByte (IOperation):
	def encode (self,value):
		return 1,chr(value)

class IOperationByteShort (IOperation):
	def encode (self,value):
		if value < (1<<8):
			return 1,chr(value)
		return 2,pack('!H',value)

# String representation for Numeric and Binary Tests

class NumericString (object):
	_string = {
		NumericOperator.LT   : '<',
		NumericOperator.GT   : '>',
		NumericOperator.EQ   : '=',
		NumericOperator.LT|NumericOperator.EQ : '<=',
		NumericOperator.GT|NumericOperator.EQ : '>=',

		NumericOperator.AND|NumericOperator.LT   : '&<',
		NumericOperator.AND|NumericOperator.GT   : '&>',
		NumericOperator.AND|NumericOperator.EQ   : '&=',
		NumericOperator.AND|NumericOperator.LT|NumericOperator.EQ : '&<=',
		NumericOperator.AND|NumericOperator.GT|NumericOperator.EQ : '&>=',
	}

	def __str__ (self):
		return "%s%s" % (self._string[self.operations & (CommonOperator.EOL ^ 0xFF) ], self.value)

	def __repr__ (self):
		return str(self)

class BinaryString (object):
	_string = {
		BinaryOperator.NOT   : '!',
		BinaryOperator.MATCH : '=',
		BinaryOperator.AND|BinaryOperator.NOT   : '&!',
		BinaryOperator.AND|BinaryOperator.MATCH : '&=',
	}

	def __str__ (self):
		return "%s%s" % (self._string[self.operations & (CommonOperator.EOL ^ 0xFF) ], self.value)

	def __repr__ (self):
		return str(self)

# Components ..............................

# Prefix
class Destination (IPrefix):
	ID = 0x01
	NAME = 'destination'

# Prefix
class Source (IPrefix):
	ID = 0x02
	NAME = 'source'

# NumericOperator
class IPProtocol (IOperationByte,NumericString):
	ID  = 0x03
	NAME = 'protocol'

# NumericOperator
class AnyPort (IOperationByteShort,NumericString):
	ID  = 0x04
	NAME = 'port'

# NumericOperator
class DestinationPort (IOperationByteShort,NumericString):
	ID  = 0x05
	NAME = 'destination-port'

# NumericOperator
class SourcePort (IOperationByteShort,NumericString):
	ID  = 0x06
	NAME = 'source-port'

# BinaryOperator
class ICMPType (IOperationByte,BinaryString):
	ID = 0x07
	NAME = 'icmp-type'

# BinaryOperator
class ICMPCode (IOperationByte,BinaryString):
	ID = 0x08
	NAME = 'icmp-code'

# BinaryOperator
class TCPFlag (IOperationByte,BinaryString):
	ID = 0x09
	NAME = 'tcp-flags'

# NumericOperator
class PacketLength (IOperationByteShort,NumericString):
	ID = 0x0A
	NAME = 'packet-length'

# NumericOperator
# RFC2474
class DSCP (IOperationByteShort,NumericString):
	ID = 0x0B
	NAME = 'dscp'

# BinaryOperator
class Fragment (IOperationByteShort,NumericString):
	ID = 0x0D
	NAME = 'fragment'

# ..........................................................

class _FlowNLRI (Attributes,Address):
	def __init__ (self,afi,safi):
		Attributes.__init__(self)
		Address.__init__(self,afi,safi)
		self.rules = {}

	def add_and (self,rule):
		ID = rule.ID
		if self.rules.has_key(ID):
			rule.first = False
			# Source and Destination do not use operations, it is just here to make the code simpler
			self.rules[ID][-1].operations |= CommonOperator.AND
		self.rules.setdefault(ID,[]).append(rule)
		return True

	def add_or (self,rule):
		ID = rule.ID
		# This test currently always fails (we do not call add_or with Source/Destinations).
		if ID in [Destination.ID, Source.ID]:
			return False
		if self.rules.has_key(ID):
			rule.first = False
		self.rules.setdefault(ID,[]).append(rule)
		return True

	def pack (self):
		ordered_rules = []

		# the order is a RFC requirement
		IDS = self.rules.keys()
		IDS.sort()

		for ID in IDS:
			rules = self.rules[ID]
			# for each component get all the operation to do
			# the format use does not prevent two opposing rules meaning that no packet can ever match
			for rule in rules:
				# clear the EOL if it has been set (it should not have been done.)
				rule.operations &= (CommonOperator.EOL ^ 0xFF)
			# and add it to the last rule
			rules[-1].operations |= CommonOperator.EOL
			for rule in rules:
				ordered_rules.append(rule)

		components = ''.join([rule.pack() for rule in ordered_rules])
		l = len(components)
		if l < 0xF0:
			data = "%s%s" % (chr(l),components)
		elif l < 0x0FFF:
			data = "%s%s" % (pack('!H',l | 0xF000),components)
		else:
			logger.critical("rule too big for NLRI - how to handle this - does this work ?")
			data = "%s" % chr(0)
		return data

	def __str__ (self):
		string = []
		for _,rules in self.rules.iteritems():
			s = []
			for rule in rules:
				if rule.operations & NumericOperator.AND:
					s.append(str(rule))
				else:
					s.append(' ')
					s.append(str(rule))
			string.append('%s %s' % (rules[0].NAME,''.join(s[1:])))
		return ' '.join(string)

	def __repr__ (self):
		return str(self)

class Flow (object):
	def __init__ (self,afi=AFI.ipv4,safi=SAFI.flow_ipv4):
		self.attributes = Attributes()
		self.nlri = _FlowNLRI(afi,safi)
		self.attributes[AttributeID.EXTENDED_COMMUNITY] = ECommunities()

	def add_and (self,rule):
		return self.nlri.add_and(rule)

	def add_or (self,rule):
		return self.nlri.add_or(rule)

	def add_action (self,community):
		self.attributes[AttributeID.EXTENDED_COMMUNITY].add(community)

	def __str__ (self):
		return "%s %s%s" % (Address.__str__(self.nlri),str(self.nlri),str(self.attributes))

	def __repr__ (self):
		return str(self)



# encoding: utf-8
"""
protocol.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

#import os
#import copy
import time
import socket
from struct import unpack

from bagpipe.exabgp.rib.table import Table
from bagpipe.exabgp.rib.delta import Delta

from bagpipe.exabgp.utils                import hexa
from bagpipe.exabgp.structure.address    import AFI,SAFI
from bagpipe.exabgp.structure.ip         import BGPPrefix,Inet,to_IP
from bagpipe.exabgp.structure.vpn        import VPNLabelledPrefix
from bagpipe.exabgp.structure.evpn       import EVPNNLRI
from bagpipe.exabgp.structure.rtc        import RouteTargetConstraint
from bagpipe.exabgp.structure.asn        import ASN,AS_TRANS
from bagpipe.exabgp.network.connection   import Connection
from bagpipe.exabgp.message              import Message,defix,Failure
from bagpipe.exabgp.message.nop          import NOP
from bagpipe.exabgp.message.open         import Open,Unknown,Parameter,Capabilities,RouterID,MultiProtocol,RouteRefresh,CiscoRouteRefresh,MultiSession,Graceful
from bagpipe.exabgp.message.update       import Update
from bagpipe.exabgp.message.update.eor   import EOR
from bagpipe.exabgp.message.keepalive    import KeepAlive
from bagpipe.exabgp.message.notification import Notification, Notify #, NotConnected
from bagpipe.exabgp.message.update.route import ReceivedRoute # ,Route
from bagpipe.exabgp.message.update.attributes     import Attributes
from bagpipe.exabgp.message.update.attribute      import AttributeID
from bagpipe.exabgp.message.update.attribute.flag        import Flag
from bagpipe.exabgp.message.update.attribute.origin      import Origin
from bagpipe.exabgp.message.update.attribute.aspath      import ASPath,AS4Path
from bagpipe.exabgp.message.update.attribute.nexthop     import NextHop
from bagpipe.exabgp.message.update.attribute.med         import MED
from bagpipe.exabgp.message.update.attribute.localpref   import LocalPreference
from bagpipe.exabgp.message.update.attribute.communities import Community,Communities,ECommunity,ECommunities
from bagpipe.exabgp.message.update.attribute.originator_id import OriginatorId
from bagpipe.exabgp.message.update.attribute.pmsi_tunnel import PMSITunnel 
#from bagpipe.exabgp.message.update.attribute.mprnlri     import MPRNLRI
#from bagpipe.exabgp.message.update.attribute.mpurnlri    import MPURNLRI
from bagpipe.exabgp.processes  import ProcessError

from bagpipe.exabgp.log import Logger
logger = Logger()

MAX_BACKLOG = 200000

# README: Move all the old packet decoding in another file to clean up the includes here, as it is not used anyway

class Protocol (object):
	decode = True
	strict = False

	def __init__ (self,peer,connection=None):
		self.peer = peer
		self.neighbor = peer.neighbor
		self.connection = connection
		self._delta = Delta(Table(peer))
		self._asn4 = False
		self._messages = {}
		self._frozen = 0
		self.message_size = 4096

	# XXX: we use self.peer.neighbor.peer_address when we could use self.neighbor.peer_address

	def me (self,message):
		return "Peer %15s ASN %-7s %s" % (self.peer.neighbor.peer_address,self.peer.neighbor.peer_as,message)

	def connect (self):
		# allows to test the protocol code using modified StringIO with a extra 'pending' function
		if not self.connection:
			peer = self.neighbor.peer_address
			local = self.neighbor.local_address
			md5 = self.neighbor.md5
			ttl = self.neighbor.ttl
			self.connection = Connection(peer,local,md5,ttl)

			message = 'neighbor %s connected\n' % self.peer.neighbor.peer_address
			try:
				proc = self.peer.supervisor.processes
				for name in proc.notify(self.neighbor.peer_address):
					proc.write(name,message)
			except ProcessError:
				raise Failure('Could not send message(s) to helper program(s) : %s' % message)

	def check_keepalive (self):
		left = int (self.connection.last_read  + self.neighbor.hold_time - time.time())
		if left <= 0:
			raise Notify(4,0)
		return left

	def close (self):
		#self._delta.last = 0
		if self.connection:
			# must be first otherwise we could have a loop caused by the raise in the below
			self.connection.close()
			self.connection = None

			message = 'neighbor %s down\n' % self.peer.neighbor.peer_address
			try:
				proc = self.peer.supervisor.processes
				for name in proc.notify(self.neighbor.peer_address):
					proc.write(name,message)
			except ProcessError:
				raise Failure('Could not send message(s) to helper program(s) : %s' % message)

	# Read from network .......................................................

	def read_message (self):
		# This call reset the time for the timeout in
		if not self.connection.pending(True):
			return NOP('')

		length = 19
		data = ''
		while length:
			if self.connection.pending():
				delta = self.connection.read(length)
				data += delta
				length -= len(delta)
				# The socket is closed
				if not data:
					raise Failure('The TCP connection is closed')

		if data[:16] != Message.MARKER:
			# We are speaking BGP - send us a valid Marker
			raise Notify(1,1,'The packet received does not contain a BGP marker')

		raw_length = data[16:18]
		length = unpack('!H',raw_length)[0]
		msg = data[18]

		if ( length < 19 or length > 4096):
			# BAD Message Length
			raise Notify(1,2)

		if (
			(msg == Open.TYPE and length < 29) or
			(msg == Update.TYPE and length < 23) or
			(msg == Notification.TYPE and length < 21) or
			(msg == KeepAlive.TYPE and length != 19)
		):
			# MUST send the faulty length back
			raise Notify(1,2,raw_length)
			#(msg == RouteRefresh.TYPE and length != 23)

		length -= 19
		data = ''
		while length:
			if self.connection.pending():
				delta = self.connection.read(length)
				data += delta
				length -= len(delta)
				# The socket is closed
				if not data:
					raise Failure('The TCP connection is closed')

		if msg == Notification.TYPE:
			raise Notification(ord(data[0]),ord(data[1]))

		if msg == KeepAlive.TYPE:
			return self.KeepAliveFactory(data)

		if msg == Open.TYPE:
			return self.OpenFactory(data)

		if msg == Update.TYPE:
			if self.neighbor.parse_routes:
				update = self.UpdateFactory(data)
				return update
			else:
				return NOP('')

		if self.strict:
			raise Notify(1,3,msg)

		return NOP(data)

	def read_open (self,_open,ip):
		message = self.read_message()

		if message.TYPE == NOP.TYPE:
			return message

		if message.TYPE != Open.TYPE:
			raise Notify(5,1,'The first packet received is not an open message (%s)' % message)

		if _open.asn.asn4() and not message.capabilities.announced(Capabilities.FOUR_BYTES_ASN):
			raise Notify(2,0,'We have an ASN4 and you do not speak it. bye.')

		self._asn4 = message.capabilities.announced(Capabilities.FOUR_BYTES_ASN)

		if message.asn == AS_TRANS:
			peer_as = message.capabilities[Capabilities.FOUR_BYTES_ASN]
		else:
			peer_as = message.asn

		if peer_as != self.neighbor.peer_as:
			raise Notify(2,2,'ASN in OPEN (%d) did not match ASN expected (%d)' % (message.asn,self.neighbor.peer_as))

		# RFC 6286 : http://tools.ietf.org/html/rfc6286
		#if message.router_id == RouterID('0.0.0.0'):
		#	message.router_id = RouterID(ip)
		if message.router_id == RouterID('0.0.0.0'):
			raise Notify(2,3,'0.0.0.0 is an invalid router_id according to RFC6286')
		if message.router_id == self.neighbor.router_id and message.asn == self.neighbor.local_as:
			raise Notify(2,3,'BGP Indendifier collision (%s) on IBGP according to RFC 6286' % message.router_id)

		if message.hold_time < 3:
			raise Notify(2,6,'Hold Time is invalid (%d)' % message.hold_time)
		if message.hold_time >= 3:
			self.neighbor.hold_time = min(self.neighbor.hold_time,message.hold_time)

		# XXX: Does not work as the capa is not yet defined
		if message.capabilities.announced(Capabilities.EXTENDED_MESSAGE):
			# untested !
			if self.peer.bgp.message_size:
				self.message_size = self.peer.bgp.message_size

# README: This limit what we are announcing may cause some issue if you add new family and SIGHUP
# README: So it is commented until I make my mind to add it or not (as Juniper complain about mismatch capabilities)
#		# Those are the capacity we need to announce those routes
#		for family in _open.capabilities[Capabilities.MULTIPROTOCOL_EXTENSIONS]:
#			# if the peer does not support them, tear down the session
#			if family not in message.capabilities[Capabilities.MULTIPROTOCOL_EXTENSIONS]:
#				afi,safi = family
#				raise Notify(2,0,'Peers does not speak %s %s' % (afi,safi))

		return message

	def read_keepalive (self):
		message = self.read_message()
		if message.TYPE == NOP.TYPE:
			return message
		if message.TYPE != KeepAlive.TYPE:
			raise Notify(5,2)
		return message

	# Sending message to peer .................................................

	# we do not buffer those message in purpose

	def new_open (self,restarted,asn4):
		if asn4:
			asn = self.neighbor.local_as
		else:
			asn = AS_TRANS

		o = Open(4,asn,self.neighbor.router_id.ip,Capabilities().default(self.neighbor,restarted),self.neighbor.hold_time)

		if not self.connection.write(o.message()):
			raise Failure('Could not send open')
		return o

	def new_keepalive (self,force=False):
		left = int(self.connection.last_write + self.neighbor.hold_time.keepalive() - time.time())
		k = KeepAlive()
		m = k.message()
		if force:
			written = self.connection.write(k.message())
			if not written:
				logger.message(self.me(">> KEEPALIVE buffered"))
				self._messages[self.neighbor.peer_as].append(('KEEPALIVE',m))
			else:
				self._frozen = 0
			return left,k
		if left <= 0:
			written = self.connection.write(k.message())
			if not written:
				logger.message(self.me(">> KEEPALIVE buffered"))
				self._messages[self.neighbor.peer_as].append(('KEEPALIVE',m))
			else:
				self._frozen = 0
			return left,k
		return left,None

	def new_notification (self,notification):
		return self.connection.write(notification.message())

	# messages buffered in case of failure

	def buffered (self):
		return self._messages.get(self.neighbor.peer_as,[]) != []

	def _backlog (self,maximum=0):
		backlog = self._messages.get(self.neighbor.peer_as,[])
		if backlog:
			if not self._frozen:
				self._frozen = time.time()
			if self._frozen and self._frozen + (self.neighbor.hold_time) < time.time():
				raise Failure('peer %s not reading on socket - killing session' % self.neighbor.peer_as)
			logger.message(self.me("unable to send route for %d second (maximum allowed %d)" % (time.time()-self._frozen,self.neighbor.hold_time)))
			nb_backlog = len(backlog)
			if nb_backlog > MAX_BACKLOG:
				raise Failure('over %d routes buffered for peer %s - killing session' % (MAX_BACKLOG,self.neighbor.peer_as))
			logger.message(self.me("backlog of %d/%d routes" % (nb_backlog,MAX_BACKLOG)))
		count = 0
		while backlog:
			count += 1
			name,update = backlog[0]
			written = self.connection.write(update)
			if not written:
				break
			logger.message(self.me(">> DEBUFFERED %s" % name))
			backlog.pop(0)
			self._frozen = 0
			yield count
			if maximum and count >= maximum:
				break
		self._messages[self.neighbor.peer_as] = backlog

	def _announce (self,name,generator):
		def chunked (generator,size):
			chunk = ''
			for data in generator:
				if len(data) > size:
					raise Failure('Can not send BGP update larger than %d bytes on this connection.' % size)
				if len(chunk) + len(data) <= size:
					chunk += data
					continue
				yield chunk
				chunk = data
			if chunk:
				yield chunk

		count = 0
		# The message size is the whole BGP message INCLUDING headers !
		for update in chunked(generator,self.message_size-19):
			count += 1
			if self._messages[self.neighbor.peer_as]:
				logger.message(self.me(">> %s could not be sent, some messages are still in the buffer" % name))
				self._messages[self.neighbor.peer_as].append((name,update))
				continue
			written = self.connection.write(update)
			if not written:
				logger.message(self.me(">> %s buffered" % name))
				self._messages[self.neighbor.peer_as].append((name,update))
			yield count

	def new_announce (self):
		for answer in self._backlog():
			yield answer
		asn4 = not not self.peer.open.capabilities.announced(Capabilities.FOUR_BYTES_ASN)
		for answer in self._announce('UPDATE',self._delta.announce(asn4,self.neighbor.local_as,self.neighbor.peer_as)):
			yield answer

	def new_update (self):
		for answer in self._backlog():
			yield answer
		asn4 = not not self.peer.open.capabilities.announced(Capabilities.FOUR_BYTES_ASN)
		for answer in self._announce('UPDATE',self._delta.update(asn4,self.neighbor.local_as,self.neighbor.peer_as)):
			yield answer

	def new_eors (self,families):
		for answer in self._backlog():
			pass
		eor = EOR()
		eors = eor.eors(families)
		for answer in self._announce('EOR',eors):
			pass

	# Message Factory .................................................

	def KeepAliveFactory (self,data):
		return KeepAlive()

	def _key_values (self,name,data):
		if len(data) < 2:
			raise Notify(2,0,"Bad length for OPEN %s (<2) %s" % (name,hexa(data)))
		l = ord(data[1])
		boundary = l+2
		if len(data) < boundary:
			raise Notify(2,0,"Bad length for OPEN %s (buffer underrun) %s" % (name,hexa(data)))
		key = ord(data[0])
		value = data[2:boundary]
		rest = data[boundary:]
		return key,value,rest

	def CapabilitiesFactory (self,data):
		capabilities = Capabilities()
		option_len = ord(data[0])
		if option_len:
			data = data[1:]
			while data:
				key,value,data = self._key_values('parameter',data)
				# Paramaters must only be sent once.
				if key == Parameter.AUTHENTIFICATION_INFORMATION:
					raise Notify(2,5)

				if key == Parameter.CAPABILITIES:
					while value:
						k,capv,value = self._key_values('capability',value)
						# Multiple Capabilities can be present in a single attribute
						#if r:
						#	raise Notify(2,0,"Bad length for OPEN %s (size mismatch) %s" % ('capability',hexa(value)))

						if k == Capabilities.MULTIPROTOCOL_EXTENSIONS:
							if k not in capabilities:
								capabilities[k] = MultiProtocol()
							afi = AFI(unpack('!H',capv[:2])[0])
							safi = SAFI(ord(capv[3]))
							capabilities[k].append((afi,safi))
							continue

						if k == Capabilities.GRACEFUL_RESTART:
							restart = unpack('!H',capv[:2])[0]
							restart_flag = restart >> 12
							restart_time = restart & Graceful.TIME_MASK
							value_gr = capv[2:]
							families = []
							while value_gr:
								afi = AFI(unpack('!H',value_gr[:2])[0])
								safi = SAFI(ord(value_gr[2]))
								flag_family = ord(value_gr[0])
								families.append((afi,safi,flag_family))
								value_gr = value_gr[4:]
							capabilities[k] = Graceful(restart_flag,restart_time,families)
							continue

						if k == Capabilities.FOUR_BYTES_ASN:
							capabilities[k] = ASN(unpack('!L',capv[:4])[0])
							continue

						if k == Capabilities.ROUTE_REFRESH:
							capabilities[k] = RouteRefresh()
							continue

						if k == Capabilities.CISCO_ROUTE_REFRESH:
							capabilities[k] = CiscoRouteRefresh()
							continue

						if k == Capabilities.MULTISESSION_BGP:
							capabilities[k] = MultiSession()
							continue
						if k == Capabilities.MULTISESSION_BGP_RFC:
							capabilities[k] = MultiSession()
							continue

						if k not in capabilities:
							capabilities[k] = Unknown(k,[ord(_) for _ in capv])
				else:
					raise Notify(2,0,'Unknown OPEN parameter %s' % hex(key))
		return capabilities

	def OpenFactory (self,data):
		version = ord(data[0])
		if version != 4:
			# Only version 4 is supported nowdays..
			raise Notify(2,1,data[0])
		asn = unpack('!H',data[1:3])[0]
		hold_time = unpack('!H',data[3:5])[0]
		numeric = unpack('!L',data[5:9])[0]
		router_id = "%d.%d.%d.%d" % (numeric>>24,(numeric>>16)&0xFF,(numeric>>8)&0xFF,numeric&0xFF)
		capabilities = self.CapabilitiesFactory(data[9:])
		return Open(version,asn,router_id,capabilities,hold_time)


	def UpdateFactory (self,data):
		length = len(data)
		# withdrawn
		lw,withdrawn,data = defix(data)
		if len(withdrawn) != lw:
			raise Notify(3,1)
		la,attribute,announced = defix(data)
		if len(attribute) != la:
			raise Notify(3,1)
		# The RFC check ...
		#if lw + la + 23 > length:
		if 2 + lw + 2+ la + len(announced) != length:
			raise Notify(3,1)

		routes = []
		while withdrawn:
			nlri = BGPPrefix(AFI.ipv4,withdrawn)
			route = ReceivedRoute(nlri,'withdraw')
			withdrawn = withdrawn[len(nlri):]
			routes.append(route)

		self.mp_routes = []
		attributes = self.AttributesFactory(attribute)
		routes.extend(self.mp_routes)

		while announced:
			nlri = BGPPrefix(AFI.ipv4,announced)
			route = ReceivedRoute(nlri,'announce')
			# XXX: Should this be a deep copy
			route.attributes = attributes
			announced = announced[len(nlri):]
			routes.append(route)
			#logger.info(self.me('Received route %s' % nlri))

		#print "routes", routes
		#print "attributes", attributes

		if routes:
			return Update(routes)
		return NOP('')

	def AttributesFactory (self,data):
		try:
			self.attributes = Attributes()
			return self._AttributesFactory(data).attributes
		except IndexError:
			raise Notify(3,2,data)

	def __new_ASPath (self,data,asn4=False):
		
		if len(data) == 0:
			return ASPath(asn4)
		
		if asn4:
			size = 4
			decoder = 'L' # could it be 'I' as well ?
		else:
			size = 2
			decoder = 'H'
		stype = ord(data[0])
		slen = ord(data[1])
		sdata = data[2:2+(slen*size)]
		
		ASPS = ASPath(asn4,stype)
		format = '!'+(decoder*slen)
		for c in unpack(format,sdata):
			ASPS.add(c)
		return ASPS

	def __new_AS4Path (self,data):
		stype = ord(data[0])
		slen = ord(data[1])
		sdata = data[2:2+(slen*4)]

		ASPS = AS4Path(stype)
		format = '!'+('L'*slen)
		for c in unpack(format,sdata):
			ASPS.add(c)
		return ASPS

	def __merge_attributes (self):
		as2path = self.attributes[AttributeID.AS_PATH]
		as4path = self.attributes[AttributeID.AS4_PATH]
		newASPS = ASPath(True,as2path.asptype)
		len2 = len(as2path.aspsegment)
		len4 = len(as4path.aspsegment)

		if len2 < len4:
			for asn in as4path.aspsegment:
				newASPS.add(asn)
		else:
			for asn in as2path.aspsegment[:-len4]:
				newASPS.add(asn)
			for asn in as4path.aspsegment:
				newASPS.add(asn)

		self.attributes.remove(AttributeID.AS_PATH)
		self.attributes.remove(AttributeID.AS4_PATH)
		self.attributes.add(newASPS)

		#raise Notify(3,1,'could not merge AS4_PATH in AS_PATH')

	def __new_communities (self,data):
		communities = Communities()
		while data:
			community = unpack('!L',data[:4])[0]
			data = data[4:]
			if data and len(data) < 4:
				raise Notify(3,1,'could not decode community %s' % str([hex(ord(_)) for _ in data]))
			communities.add(Community(community))
		return communities

	def __new_extended_communities (self,data):
		communities = ECommunities()
		while data:
			community = data[:8]
			data = data[8:]
			if data and len(data) < 8:
				raise Notify(3,1,'could not decode extended community %s' % str([hex(ord(_)) for _ in data]))
			communities.add(ECommunity.unpackFrom(community))
		return communities

	def _AttributesFactory (self,data):
		if not data:
			return self

		# We do not care if the attribute are transitive or not as we do not redistribute
		flag = Flag(ord(data[0]))
		code = AttributeID(ord(data[1]))

		if flag & Flag.EXTENDED_LENGTH:
			length = unpack('!H',data[2:4])[0]
			offset = 4
		else:
			length = ord(data[2])
			offset = 3

		data = data[offset:]

#		if not length:
#			return self._AttributesFactory(data[length:])

		# XXX: This code does not make sure that attributes are unique - or does it ?

		if code == AttributeID.ORIGIN:
			logger.parser('parsing origin')
			self.attributes.add(Origin(ord(data[0])))
			return self._AttributesFactory(data[length:])

		if code == AttributeID.AS_PATH:
			logger.parser('parsing as_path')
			self.attributes.add(self.__new_ASPath(data[:length],self._asn4))
			if not self._asn4 and self.attributes.has(AttributeID.AS4_PATH):
				self.__merge_attributes()
			return self._AttributesFactory(data[length:])

		if code == AttributeID.AS4_PATH:
			logger.parser('parsing as_path')
			self.attributes.add(self.__new_AS4Path(data[:length]))
			if not self._asn4 and self.attributes.has(AttributeID.AS_PATH):
				self.__merge_attributes()
			return self._AttributesFactory(data[length:])

		if code == AttributeID.NEXT_HOP:
			logger.parser('parsing next-hop')
			self.attributes.add(NextHop(Inet(AFI.ipv4,data[:4])))
			return self._AttributesFactory(data[length:])

		if code == AttributeID.MED:
			logger.parser('parsing med')
			self.attributes.add(MED(unpack('!L',data[:4])[0]))
			return self._AttributesFactory(data[length:])

		if code == AttributeID.LOCAL_PREF:
			logger.parser('parsing local-preference')
			self.attributes.add(LocalPreference(unpack('!L',data[:4])[0]))
			return self._AttributesFactory(data[length:])
		
		if code == AttributeID.ORIGINATOR_ID:
			logger.parser('parsing originator-id')
			self.attributes.add(OriginatorId.unpack(data[:4]))
			return self._AttributesFactory(data[length:])

		if code == AttributeID.PMSI_TUNNEL:
			logger.parser('parsing pmsi-tunnel')
			self.attributes.add(PMSITunnel.unpack(data[:length]))
			return self._AttributesFactory(data[length:])

		if code == AttributeID.ATOMIC_AGGREGATE:
			logger.parser('ignoring atomic-aggregate')
			return self._AttributesFactory(data[length:])

		if code == AttributeID.AGGREGATOR:
			logger.parser('ignoring aggregator')
			return self._AttributesFactory(data[length:])

		if code == AttributeID.AS4_AGGREGATOR:
			logger.parser('ignoring as4_aggregator')
			return self._AttributesFactory(data[length:])

		if code == AttributeID.COMMUNITY:
			logger.parser('parsing communities')
			self.attributes.add(self.__new_communities(data[:length]))
			return self._AttributesFactory(data[length:])

		if code == AttributeID.EXTENDED_COMMUNITY:
			logger.parser('parsing communities')
			self.attributes.add(self.__new_extended_communities(data[:length]))
			return self._AttributesFactory(data[length:])

		if code == AttributeID.MP_UNREACH_NLRI:
			logger.parser('parsing multi-protocol nlri unreacheable')
			next_attributes = data[length:]
			data = data[:length]
			afi,safi = unpack('!HB',data[:3])
			offset = 3
			# See RFC 5549 for better support
			if not afi in (AFI.ipv4,AFI.ipv6,AFI.l2vpn) or (not safi in (SAFI.unicast, SAFI.mpls_vpn, SAFI.rtc,SAFI.evpn)):
				#self.log.out('we only understand IPv4/IPv6 and should never have received this MP_UNREACH_NLRI (%s %s)' % (afi,safi))
				raise Exception("Unsupported AFI/SAFI received !! not supposed to happen here...")
				return self._AttributesFactory(next_attributes)
			data = data[offset:]
			while data:
				
				if safi == SAFI.unicast:
					route = ReceivedRoute(BGPPrefix(afi,data),'withdraw')
				elif (afi == AFI.ipv4 and safi == SAFI.mpls_vpn):
					route = ReceivedRoute(VPNLabelledPrefix.unpack(afi,safi,data),'withdraw')
				elif (afi == AFI.ipv4 and safi == SAFI.rtc):
					route = ReceivedRoute(RouteTargetConstraint.unpack(afi,safi,data),'withdraw')
				elif (afi == AFI.l2vpn and safi == SAFI.evpn):
					route = ReceivedRoute(EVPNNLRI.unpack(data) ,'withdraw')
				else:
					raise Exception("Unsupported AFI/SAFI combination !!")
					return self._AttributesFactory(next_attributes)
				
				data = data[len(route.nlri):]
				self.mp_routes.append(route)
			return self._AttributesFactory(next_attributes)

		if code == AttributeID.MP_REACH_NLRI:
			logger.parser('parsing multi-protocol nlri reacheable')
			next_attributes = data[length:]
			data = data[:length]
			afi,safi = unpack('!HB',data[:3])
			offset = 3
			
			if not afi in (AFI.ipv4,AFI.ipv6,AFI.l2vpn) or (not safi in (SAFI.unicast,SAFI.mpls_vpn,SAFI.rtc,SAFI.evpn)):
				#self.log.out('we only understand IPv4/IPv6 and should never have received this MP_REACH_NLRI (%s %s)' % (afi,safi))
				raise Exception("Unsupported AFI/SAFI received !! not supposed to happen here...")
				return self._AttributesFactory(next_attributes)
			len_nh = ord(data[offset])
			offset += 1
			if afi == AFI.ipv4 and safi in (SAFI.unicast,) and not len_nh == 4: 
				# We are not following RFC 4760 Section 7 (deleting route and possibly tearing down the session)
				#self.log.out('bad IPv4 next-hop length (%d)' % len_nh)
				return self._AttributesFactory(next_attributes)
			if afi == AFI.ipv6 and safi in (SAFI.unicast,) and not len_nh in (16,32):
				# We are not following RFC 4760 Section 7 (deleting route and possibly tearing down the session)
				#self.log.out('bad IPv6 next-hop length (%d)' % len_nh)
				return self._AttributesFactory(next_attributes)
			nh = data[offset:offset+len_nh]
			offset += len_nh
			if len_nh == 32:
				# we have a link-local address in the next-hop we ideally need to ignore
				if nh[0] == chr(0xfe): nh = nh[16:]
				elif nh[16] == chr(0xfe): nh = nh[:16]
				# We are not following RFC 4760 Section 7 (deleting route and possibly tearing down the session)
				else: return self._AttributesFactory(next_attributes)
			if len_nh >= 16: nh = socket.inet_ntop(socket.AF_INET6,nh)
			else:
				
				if (safi in (SAFI.mpls_vpn,)):
					# the next-hop is preceded by an rdtype and a 6-byte RD, we don't care about the RD yet
					nh = socket.inet_ntop(socket.AF_INET,nh[8:])
				else:
					nh = socket.inet_ntop(socket.AF_INET,nh)
			
			nb_snpa = ord(data[offset])
			offset += 1
			snpas = []
			for _ in range(nb_snpa):
				len_snpa = ord(offset)
				offset += 1
				snpas.append(data[offset:offset+len_snpa])
				offset += len_snpa
			data = data[offset:]
			while data:
				if safi == SAFI.unicast:
					route = ReceivedRoute(BGPPrefix(afi,data),'announce')
				elif (afi == AFI.ipv4 and safi == SAFI.mpls_vpn):
					route = ReceivedRoute(VPNLabelledPrefix.unpack(afi,safi,data) ,'announce')
				elif (afi == AFI.ipv4 and safi == SAFI.rtc):
					route = ReceivedRoute(RouteTargetConstraint.unpack(afi,safi,data) ,'announce')
				elif (afi == AFI.l2vpn and safi == SAFI.evpn):
					route = ReceivedRoute(EVPNNLRI.unpack(data) ,'announce')
				else:
					raise Exception("Unsupported AFI/SAFI combination !!")
					return self._AttributesFactory(next_attributes)
										
				data = data[len(route.nlri):]
				route.attributes = self.attributes
				route.attributes.add(NextHop(to_IP(nh)))
				self.mp_routes.append(route)
			return self._AttributesFactory(next_attributes)

		logger.warning("ignoring attributes of type %s %s" % (str(code),[hex(ord(_)) for _ in data]),'parsing')
		return self._AttributesFactory(data[length:])




# encoding: utf-8
"""
daemon.py

Created by Thomas Mangin on 2011-05-02.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

import os
import sys
import pwd
import errno
import socket

from bagpipe.exabgp.log import Logger
logger = Logger()

class Daemon (object):
	pid_file = os.environ.get('PID','')
	user = os.environ.get('USER','nobody')

	def __init__ (self,supervisor):
		self.supervisor = supervisor
		#mask = os.umask(0137)

	def savepid (self):
		self._saved_pid = False

		if not self.pid_file:
			return

		ownid = os.getpid()

		flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY
		mode = ((os.R_OK | os.W_OK) << 6) | (os.R_OK << 3) | os.R_OK

		try:
			fd = os.open(self.pid_file,flags,mode)
		except OSError:
			logger.daemon("PIDfile already exists, not updated %s" % self.pid_file)
			return

		try:
			f = os.fdopen(fd,'w')
			line = "%d\n" % ownid
			f.write(line)
			f.close()
			self._saved_pid = True
		except IOError:
			logger.daemon("Can not create PIDfile %s" % self.pid_file)
			return
		logger.daemon("Created PIDfile %s with value %d" % (self.pid_file,ownid))

	def removepid (self):
		if not self.pid_file or not self._saved_pid:
			return
		try:
			os.remove(self.pid_file)
		except OSError, e:
			if e.errno == errno.ENOENT:
				pass
			else:
				logger.daemon("Can not remove PIDfile %s" % self.pid_file)
				return
		logger.daemon("Removed PIDfile %s" % self.pid_file)

	def drop_privileges (self):
		"""returns true if we are left with insecure privileges"""
		# os.name can be ['posix', 'nt', 'os2', 'ce', 'java', 'riscos']
		if os.name not in ['posix',]:
			return False

		uid = os.getpid()
		gid = os.getgid()

		if uid and gid:
			return False

		try:
			user = pwd.getpwnam(self.user)
			nuid = int(user.pw_uid)
			ngid = int(user.pw_uid)
		except KeyError:
			return True

		# not sure you can change your gid if you do not have a pid of zero
		try:
			if not uid:
				os.setuid(nuid)
			if not gid:
				os.setgid(ngid)
			return False
		except OSError:
			return True

	def _is_socket (self,fd):
		try:
			s = socket.fromfd(fd, socket.AF_INET, socket.SOCK_RAW)
		except ValueError,e:
			# The file descriptor is closed
			return False
		try:
			s.getsockopt(socket.SOL_SOCKET, socket.SO_TYPE)
		except socket.error, e:
			# It is look like one but it is not a socket ...
			if e.args[0] == errno.ENOTSOCK:
				return False
		return True

	def daemonise (self):
		if os.environ.get('DAEMONIZE','0') not in ['','1','yes','Yes','YES']:
			return

		def fork_exit ():
			try:
				pid = os.fork()
				if pid > 0:
					os._exit(0)
			except OSError, e:
				logger.critial('Can not fork, errno %d : %s' % (e.errno,e.strerror),'supervisor')

		# do not detach if we are already supervised or run by init like process
		if not self._is_socket(sys.__stdin__.fileno()) and os.getppid() != 1:
			fork_exit()
			os.setsid()
			fork_exit()
			if os.environ.get('SYSLOG',None) is None:
				# XXX: Force SYSLOG, we need to be able to access the local syslog deamon
				# XXX: Is there not a bug on python (old version) where that does not work ?
				os.environ['SYSLOG']=''



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

import types

import traceback

import logging

import socket

from bagpipe.bgp.engine.worker import Worker
from bagpipe.bgp.engine import RouteEvent, RouteEntry

from bagpipe.bgp.common.looking_glass import LookingGlassLocalLogger, LGMap

from bagpipe.bgp.common.utils import plural
from bagpipe.bgp.common import logDecorator

from bagpipe.exabgp.message.update.attribute import AttributeID
from bagpipe.exabgp.message.update.attributes import Attributes

keepAttributes_default = [AttributeID.NEXT_HOP,
                          AttributeID.PMSI_TUNNEL,
                          AttributeID.MED,
                          AttributeID.EXTENDED_COMMUNITY,  # FIXME
                          AttributeID.LOCAL_PREF]


class FilteredRouteEntry(RouteEntry):

    def __init__(self, re, keepAttributes=None):
        if keepAttributes is None:
            keepAttributes = keepAttributes_default

        attributes = Attributes()
        for (attributeId, attribute) in re.attributes.iteritems():
            if attributeId in keepAttributes:
                attributes.add(attribute)

        RouteEntry.__init__(
            self, re.afi, re.safi, re.routeTargets, re.nlri, attributes, None)


def filteredRoutes(routes):
    return [FilteredRouteEntry(route) for route in routes]


# def _compareRoutes(self, routeA, routeB):
#         """
#         should return:
#          - an int>0 if routeA is better than routeB
#          - an int<0 if routeB is better than routeA
#          - else 0
#         """

# TODO: both comparison should first compare local_pref and MAC Mobility
# if present


def compareNoECMP(self, routeA, routeB):
    '''
    This compares the two routes in a consistent fashion, but two routes
    will never be considered of equal cost.
    The comparison is 'salted' so that two distinct VRFs (e.g. on two distinct
    bagpipe-bgp instances will not necessarily elect the same route as the
    best one.
    '''
    self.log.debug("compareNoECMP used")
    salt = socket.gethostname() + self.name
    return cmp([hash(salt + repr(routeA)), routeA],
               [hash(salt + repr(routeB)), routeB])


def compareECMP(self, routeA, routeB):
    self.log.debug("compareECMP used")
    return 0


class TrackerWorker(Worker, LookingGlassLocalLogger):
    __metaclass__ = ABCMeta

    def __init__(self, bgpManager, workerName, compareRoutes=compareNoECMP):
        Worker.__init__(self, bgpManager, workerName)
        LookingGlassLocalLogger.__init__(self)

        # dict: entry -> list of routes:
        self.trackedEntry2routes = dict()
        # dict: entry -> set of bestRoutes:
        self.trackedEntry2bestRoutes = dict()

        self._compareRoutes = compareRoutes

    def getBestRoutesForTrackedEntry(self, entry):
        return self.trackedEntry2bestRoutes.get(entry, set())

    @logDecorator.log
    def _onEvent(self, routeEvent):
        newRoute = routeEvent.routeEntry
        filteredNewRoute = FilteredRouteEntry(newRoute)

        entry = self._route2trackedEntry(newRoute)

        self.log.debug("trackedEntry for this route: %s (type: %s)",
                       TrackerWorker._displayEntry(entry), type(entry))

        self._dumpState()

        try:
            allRoutes = self.trackedEntry2routes[entry]
        except KeyError:
            self.log.debug("Initiating trackedEntry2routes[entry]")
            allRoutes = []
            self.trackedEntry2routes[entry] = allRoutes

        self.log.debug("We currently have %d route%s for this entry",
                       len(allRoutes), plural(allRoutes))

        if routeEvent.type == RouteEvent.ADVERTISE:

            withdrawnBestRoutes = []

            try:
                bestRoutes = self.trackedEntry2bestRoutes[entry]

                if routeEvent.replacedRoute is not None:
                    self.log.debug("Removing replaced route from allRoutes"
                                   " and bestRoutes: %s",
                                   routeEvent.replacedRoute)
                    try:
                        allRoutes.remove(routeEvent.replacedRoute)
                    except ValueError:
                        # we did not have any route for this entry
                        self.log.error("replacedRoute is an entry for which "
                                       "we had no route ??? (bug ?)")

                    if routeEvent.replacedRoute in bestRoutes:
                        self.log.debug(
                            "Removing replacedRoute from bestRoutes")
                        bestRoutes.remove(routeEvent.replacedRoute)
                        withdrawnBestRoutes.append(routeEvent.replacedRoute)
                    else:
                        self.log.debug("replacedRoute is not in bestRoutes")
                else:
                    self.log.debug("No replaced route to remove")

                callNewBestRoute4All = False
                if len(bestRoutes) == 0:
                    self.log.debug("All best routes have been replaced")
                    self._recomputeBestRoutes(allRoutes, bestRoutes)
                    if bestRoutes:
                        currentBestRoute = iter(bestRoutes).next()
                        self.log.debug("We'll need to call newBestRoute for "
                                       "all our new best routes")
                        callNewBestRoute4All = True
                    else:
                        currentBestRoute = None
                        callNewBestRoute4All = False
                else:
                    # (if there is more than one route in the best routes, we
                    # take the first one)
                    currentBestRoute = iter(bestRoutes).next()

                    self.log.debug("Current best route: %s", currentBestRoute)

                    if newRoute == currentBestRoute:
                        self.log.info("New route is a route we already had, "
                                      "nothing to do.")
                        # nothing to do
                        return

                # let's find if we need to update our best routes
                if currentBestRoute:
                    routeComparison = self._compareRoutes(self, newRoute,
                                                          currentBestRoute)
                else:
                    routeComparison = 1

                self.log.debug("routeComparison: %d", routeComparison)

                if routeComparison > 0:
                    # newRoute is a strictly better route than any current
                    # one, discard all the current best routes
                    self.log.debug("Replacing all best routes with new one")
                    withdrawnBestRoutes.extend(bestRoutes.copy())
                    bestRoutes.clear()
                    bestRoutes.add(newRoute)
                    self._callNewBestRoute(entry, filteredNewRoute)
                    callNewBestRoute4All = False
                elif routeComparison == 0:
                    # newRoute is as good as the current ones
                    self.log.debug("Adding newRoute to bestRoutes...")

                    if callNewBestRoute4All:
                        self._callNewBestRouteForRoutes(entry, bestRoutes)

                    # We'll do a call to self._newBestRoute... *only* if the
                    # newRoute is different from all current best routes. This
                    # comparison uses FilteredRouteEntry to *not* take into
                    # account .source (the BGP peer which advertized the route)
                    # and only takes into account a specific set of BGP
                    # attributes.
                    # TODO: explain more on theses BGP attributes
                    #       related to the cases where a route is re-advertized
                    #       with updated attributes
                    isReallyNew = (FilteredRouteEntry(newRoute) not in
                                   filteredRoutes(bestRoutes))

                    bestRoutes.add(newRoute)

                    if isReallyNew:
                        self.log.debug("Calling self._newBestRoute since we "
                                       "yet had no such route in best routes")
                        self._callNewBestRoute(entry, filteredNewRoute)
                    else:
                        self.log.debug("Not calling _newBestRoute since we had"
                                       " received a similar route already")

                else:
                    self.log.debug("The route is no better than current "
                                   "best ones")

                    if callNewBestRoute4All:
                        self._callNewBestRouteForRoutes(entry, bestRoutes)

            except (KeyError, StopIteration) as e:
                self.log.debug("We had no route for this entry (%s)", e)
                self.trackedEntry2bestRoutes[entry] = set([newRoute])
                bestRoutes = set()
                self.log.debug("Calling newBestRoute")
                self._callNewBestRoute(entry, filteredNewRoute)

            # We need to call self._bestRouteRemoved for routes that where
            # implicitly withdrawn, but only if they don't have an equal route
            # (in the sense of FilteredRouteEntry) in bestRoutes
            filteredBestRoutes = filteredRoutes(bestRoutes)
            self.log.debug("Considering implicitly withdrawn best routes")
            for r in withdrawnBestRoutes:
                filteredR = FilteredRouteEntry(r)
                if filteredR not in filteredBestRoutes:
                    self.log.debug("   calling self._bestRouteRemoved for "
                                   "route: %s (not last)", filteredR)
                    self._callBestRouteRemoved(entry, filteredR, last=False)
                else:
                    self.log.debug("   not calling self._bestRouteRemoved for"
                                   " route: %s", filteredR)

            # add the route to the list of routes for this entry
            self.log.debug("Adding route to allRoutes for this entry")
            allRoutes.append(newRoute)

        else:  # RouteEvent.WITHDRAW

            withdrawnRoute = newRoute

            self.log.debug("Removing route from allRoutes for this entry")

            try:
                allRoutes.remove(withdrawnRoute)
            except ValueError:
                # we did not have any route for this entry
                self.log.error("Withdraw received for an entry for which we"
                               " had no route ??? (not supposed to happen)")

            try:
                bestRoutes = self.trackedEntry2bestRoutes[entry]

                if withdrawnRoute in bestRoutes:
                    self.log.debug("The event received is about a route which"
                                   " is among the best routes for this entry")
                    # remove the route from bestRoutes
                    bestRoutes.remove(withdrawnRoute)

                    withdrawnRouteIsLast = True
                    if len(bestRoutes) == 0:
                        # we don't have any best route left...
                        self._recomputeBestRoutes(allRoutes, bestRoutes)

                        if len(bestRoutes) > 0:
                            self._callNewBestRouteForRoutes(entry, bestRoutes)
                            withdrawnRouteIsLast = False
                        else:
                            self.log.debug("Cleanup allRoutes and bestRoutes")
                            del self.trackedEntry2bestRoutes[entry]
                            del self.trackedEntry2routes[entry]
                    else:
                        # we still have some best routes, no new best route
                        # call to do
                        withdrawnRouteIsLast = False

                    self.log.debug("Calling bestRouteRemoved...?")
                    # We need to call self._bestRouteRemoved, but only if the
                    # withdrawn route does not have an equal route in
                    # bestRoutes (in the sense of FilteredRouteEntry)
                    filteredWithdrawnRoute = FilteredRouteEntry(withdrawnRoute)
                    if (filteredWithdrawnRoute not
                            in filteredRoutes(bestRoutes)):
                        self.log.debug("Calling bestRouteRemoved: %s(last:%s)",
                                       filteredWithdrawnRoute,
                                       withdrawnRouteIsLast)
                        self._callBestRouteRemoved(entry,
                                                   filteredWithdrawnRoute,
                                                   withdrawnRouteIsLast)
                    else:
                        self.log.debug("No need to call bestRouteRemved: %s",
                                       filteredWithdrawnRoute)

                else:
                    self.log.debug("The event received is not related to one "
                                   "of the best routes for this entry")
                    # no need to update our best route list
                    pass

            except (KeyError, ValueError) as e:
                # we did not have any route for this entry
                self.log.error("Withdraw received for an entry for which we "
                               "had no route: not supposed to happen! (%s)", e)
            except Exception:
                raise

        self.log.info("We now have %d route%s for this entry.", len(allRoutes),
                      plural(allRoutes))

        self._dumpState()

    def _recomputeBestRoutes(self, allRoutes, bestRoutes):
        '''update bestRoutes to contain the best routes from allRoutes, based
        on _compareRoutes'''

        newBestRoutes = []
        for route in allRoutes:
            if len(newBestRoutes) == 0:
                # self.log.debug("first route, thus our current best:
                # %s",route)
                newBestRoutes = [route]
                continue

            comparison = self._compareRoutes(self, route, newBestRoutes[0])
            # self.log.debug("Route comparison: %s vs %s == %d",(route,
            # newBestRoutes[0],comparison) )
            if comparison > 0:
                # self.log.debug("better, replaces our current best: %s",route)
                newBestRoutes = [route]
            elif comparison == 0:
                # self.log.debug("as good as our current best: %s",route)
                newBestRoutes.append(route)
            else:
                # self.log.debug("no better than our current best: %s",route)
                pass

        bestRoutes.clear()
        bestRoutes.update(newBestRoutes)

        self.log.debug("Recomputed new best routes: %s", bestRoutes)

    def _callNewBestRouteForRoutes(self, entry, routes):
        self.log.debug("Calling newBestRoute for routes, without dups")
        self.log.debug("   Routes: %s", routes)
        routesNoDups = set([FilteredRouteEntry(r) for r in routes])
        self.log.debug("   After filtering duplicates: %s", routesNoDups)
        for route in routesNoDups:
            self._callNewBestRoute(entry, route)

    def _callNewBestRoute(self, entry, newRoute):
        try:
            self._newBestRoute(entry, newRoute)
        except Exception as e:
            self.log.error("Exception in <subclass>._newBestRoute: %s", e)
            if self.log.isEnabledFor(logging.WARNING):
                self.log.info("%s", traceback.format_exc())

    def _callBestRouteRemoved(self, entry, oldRoute, last):
        try:
            self._bestRouteRemoved(entry, oldRoute, last)
        except Exception as e:
            self.log.error("Exception in <subclass>._bestRouteRemoved: %s", e)
            if self.log.isEnabledFor(logging.WARNING):
                self.log.info("%s", traceback.format_exc())

    # Callbacks for subclasses ########################

    @abstractmethod
    def _route2trackedEntry(self, route):
        """
        This method is how the subclass maps a BGP route into an object that
        the TrackerWorker code will track.

        For instance, a VPN VRF is expected to keep track of IP prefixes;
        hence the route2trackedEntry code for a VRF could return the IP prefix
        in the VPNv4 route.
        The result will be that the TrackerWorker code will keep track, for a
        each prefix, of all the routes and of the best routes.
        """
        pass

    # FIXME: need to document the behavior of these callbacks

    @abstractmethod
    def _newBestRoute(self, entry, newRoute):
        pass

    @abstractmethod
    def _bestRouteRemoved(self, entry, oldRoute, last):
        pass

    # Debug support methods #########

    def _dumpState(self):
        if self.log.isEnabledFor(logging.DEBUG):
            self.log.debug("--- trackedEntry2routes ---")
            for entry in self.trackedEntry2routes:
                self.log.debug(
                    "  Entry: %s", TrackerWorker._displayEntry(entry))
                for route in self.trackedEntry2routes[entry]:
                    self.log.debug("    Route: %s", route)

            self.log.debug("--- trackedEntry2bestRoutes ---")
            for entry in self.trackedEntry2bestRoutes:
                self.log.debug(
                    "  Entry: %s", TrackerWorker._displayEntry(entry))
                for route in self.trackedEntry2bestRoutes[entry]:
                    self.log.debug("    Route: %s", route)

            self.log.debug("--- ---")

    @staticmethod
    def _displayEntry(entry):
        if (isinstance(entry, tuple) and len(entry) > 0 and
            (isinstance(entry[0], type) or
             isinstance(entry[0], types.ClassType))):
            return repr(tuple([entry[0].__name__] + list(entry[1:])))
        else:
            return repr(entry)

    # Looking glass ###########

    def getLGMap(self):
        return {"received_routes": (LGMap.SUBTREE, self.getLGAllRoutes),
                "best_routes": (LGMap.SUBTREE, self.getLGBestRoutes)}

    def getLGAllRoutes(self, pathPrefix):
        return self._getLGRoutes(pathPrefix, self.trackedEntry2routes)

    def getLGBestRoutes(self, pathPrefix):
        return self._getLGRoutes(pathPrefix, self.trackedEntry2bestRoutes)

    def _getLGRoutes(self, pathPrefix, routeDict):
        '''
        routeDict is whether self.trackedEntry2bestRoutes or
        self.trackedEntry2routes
        '''
        routes = {}
        for entry in routeDict.iterkeys():
            routes[repr(entry)] = [route.getLookingGlassInfo(pathPrefix)
                                   for route in routeDict[entry]]
        return routes



# encoding: utf-8
"""
neighbor.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

from copy import copy

from bagpipe.exabgp.structure.address import AFI
from bagpipe.exabgp.message.open import HoldTime
from bagpipe.exabgp.message.update.attribute.id import AttributeID

from bagpipe.exabgp.log import Logger
logger = Logger()

# The definition of a neighbor (from reading the configuration)
class Neighbor (object):
	def __init__ (self):
		self.description = ''
		self.router_id = None
		self.local_address = None
		self.peer_address = None
		self.peer_as = None
		self.local_as = None
		self.hold_time = HoldTime(180)
		self.graceful_restart = False
		self.md5 = None
		self.ttl = None
		self.multisession = None
		self.parse_routes = None
		self._families = {}
		self._watchdog = {}

	def name (self):
		if self.multisession:
			session =  ", ".join("%s %s" % (afi,safi) for (afi,safi) in self._families.keys())
			return "%s-%s multi-session %s" % (self.local_address,self.peer_address,session)
		else:
			return "%s-%s" % (self.local_address,self.peer_address)

	def families (self):
		return self._families.keys()

	def watchdog (self,watchdog):
		self._watchdog = copy(watchdog)

	def every_routes (self):
		for family in self._families:
			for route in self._families[family]:
				yield route

	def filtered_routes (self):
		# This function returns a hash and not a list as "in" tests are O(n) with lists and O(1) with hash
		# and with ten thousands routes this makes an enormous difference (60 seconds to 2)
		routes = {}
		for family in self._families:
			for route in self._families[family]:
				withdrawn = route.attributes.pop(AttributeID.INTERNAL_WITHDRAWN,None)
				if withdrawn is not None:
					logger.rib('skipping initial announcement of %s' % route)
					watchdog = route.attributes.get(AttributeID.INTERNAL_WATCHDOG,None)
					if watchdog in self._watchdog:
						self._watchdog[watchdog] == 'withdraw'
					continue
				watchdog = route.attributes.get(AttributeID.INTERNAL_WATCHDOG,None)
				if watchdog in self._watchdog:
					if self._watchdog[watchdog] == 'withdraw':
						continue
				routes[str(route)] = route
		return routes

	def remove_family (self,family):
		if family in self._families:
			del self._families[family]

	def add_route (self,route):
		self._families.setdefault((route.nlri.afi,route.nlri.safi),[]).append(route)

	def remove_route (self,route):
		result = False
		for r in self._families.get((route.nlri.afi,route.nlri.safi),[]):
			if r == route:
				self._families[(route.nlri.afi,route.nlri.safi)].remove(r)
				result = True
		return result

	def set_routes (self,routes):
		self._families = {}
		for route in routes:
			self.add_route(route)

	def missing (self):
		if self.local_address is None: return 'local-address'
		if self.peer_address is None: return 'peer-address'
		if self.local_as is None: return 'local-as'
		if self.peer_as is None: return 'peer-as'
		if self.peer_address.afi == AFI.ipv6 and not self.router_id: return 'router-id'
		return ''

	# This function only compares the neighbor BUT NOT ITS ROUTES
	def __eq__ (self,other):
		return \
			self.router_id == other.router_id and \
			self.local_address == other.local_address and \
			self.local_as == other.local_as and \
			self.peer_address == other.peer_address and \
			self.peer_as == other.peer_as and \
			self.hold_time == other.hold_time and \
			self.graceful_restart == other.graceful_restart and \
			self.md5 == other.md5 and \
			self.ttl == other.ttl and \
			self.multisession == other.multisession and \
			self.families() == other.families()

	def __ne__(self, other):
		return not (self == other)

	def __str__ (self):
		routes = '\n\t\t'
		for family in self._families:
			for _routes in self._families[family]:
				routes += '\n\t\t%s' % _routes

		options = []
		if self.md5: options.append("md5: %s;" % self.md5)
		if self.ttl is not None: options.append("ttl-security: %d;" % self.ttl)
		if self.graceful_restart: options.append("graceful-restart: %d;" % self.graceful_restart)

		return """\
neighbor %s {
	description "%s";
	router-id %s;
	local-address %s;
	local-as %s;
	peer-as %s;
	hold-time %s;
	%s
	static {%s
	}
}""" % (
	self.peer_address,
	self.description,
	self.router_id,
	self.local_address,
	self.local_as,
	self.peer_as,
	self.hold_time,
	'\n\t'.join(options),
	routes
)

	def __repr__ (self):
		return str(self)



# encoding: utf-8
"""
icmp.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

# =================================================================== ICMP Code Field

# http://www.iana.org/assignments/icmp-parameters
class ICMPType (int):
	ECHO_REPLY               = 0x00
	DESTINATION_UNREACHEABLE = 0x03
	SOURCE_QUENCH            = 0x04
	REDIRECT                 = 0x05
	ECHO_REQUEST             = 0x08
	ROUTER_ADVERTISEMENT     = 0x09
	ROUTER_SOLICIT           = 0x0A
	TIME_EXCEEDED            = 0x0B
	PARAMETER_PROBLEM        = 0x0C
	TIMESTAMP_REQUEST        = 0x0D # wonder why junos call all the other ones _REQUEST but not this one
	TIMESTAMP_REPLY          = 0x0E
	INFO_REQUEST             = 0x0F
	INFO_REPLY               = 0x10
	MASK_REQUEST             = 0x11
	MASK_REPLY               = 0x12
	TRACEROUTE               = 0x1E

	def __str__ (self):
		if self == ICMPType.ECHO_REPLY:               return 'echo-reply'
		if self == ICMPType.ECHO_REQUEST:             return 'echo-request'
		if self == ICMPType.INFO_REPLY:               return 'info-reply'
		if self == ICMPType.INFO_REQUEST:             return 'info-request'
		if self == ICMPType.MASK_REPLY:               return 'mask-reply'
		if self == ICMPType.MASK_REQUEST:             return 'mask-request'
		if self == ICMPType.PARAMETER_PROBLEM:        return 'parameter-problem'
		if self == ICMPType.REDIRECT:                 return 'redirect'
		if self == ICMPType.ROUTER_ADVERTISEMENT:     return 'router-advertisement'
		if self == ICMPType.ROUTER_SOLICIT:           return 'router-solicit'
		if self == ICMPType.SOURCE_QUENCH:            return 'source-quench'
		if self == ICMPType.TIME_EXCEEDED:            return 'time-exceeded'
		if self == ICMPType.TIMESTAMP_REQUEST:        return 'timestamp'
		if self == ICMPType.TIMESTAMP_REPLY:          return 'timestamp-reply'
		if self == ICMPType.DESTINATION_UNREACHEABLE: return 'unreachable'
		return 'invalid icmp type %d' % int.__str__(self)

	def __repr__ (self):
		return str(self)

def NamedICMPType (name):
	icmp = name.lower()
	if icmp == 'echo-reply':          return ICMPType.ECHO_REPLY
	if icmp == 'echo-request':        return ICMPType.ECHO_REQUEST
	if icmp == 'info-reply':          return ICMPType.INFO_REPLY
	if icmp == 'info-request':        return ICMPType.INFO_REQUEST
	if icmp == 'mask-reply':          return ICMPType.MASK_REPLY
	if icmp == 'mask-request':        return ICMPType.MASK_REQUEST
	if icmp == 'parameter-problem':   return ICMPType.PARAMETER_PROBLEM
	if icmp == 'redirect':            return ICMPType.REDIRECT
	if icmp == 'router-advertisement':return ICMPType.ROUTER_ADVERTISEMENT
	if icmp == 'router-solicit':      return ICMPType.ROUTER_SOLICIT
	if icmp == 'source-quench':       return ICMPType.SOURCE_QUENCH
	if icmp == 'time-exceeded':       return ICMPType.TIME_EXCEEDED
	if icmp == 'timestamp':           return ICMPType.TIMESTAMP_REQUEST
	if icmp == 'timestamp-reply':     return ICMPType.TIMESTAMP_REPLY
	if icmp == 'unreachable':         return ICMPType.DESTINATION_UNREACHEABLE
	raise ValueError('unknow icmp type %s' % icmp)


# http://www.iana.org/assignments/icmp-parameters
class ICMPCode (int):
	# Destination Unreacheable (type 3)
	NETWORK_UNREACHABLE                   = 0x0
	HOST_UNREACHABLE                      = 0x1
	PROTOCOL_UNREACHABLE                  = 0x2
	PORT_UNREACHABLE                      = 0x3
	FRAGMENTATION_NEEDED                  = 0x4
	SOURCE_ROUTE_FAILED                   = 0x5
	DESTINATION_NETWORK_UNKNOWN           = 0x6
	DESTINATION_HOST_UNKNOWN              = 0x7
	SOURCE_HOST_ISOLATED                  = 0x8
	DESTINATION_NETWORK_PROHIBITED        = 0x9
	DESTINATION_HOST_PROHIBITED           = 0xA
	NETWORK_UNREACHABLE_FOR_TOS           = 0xB
	HOST_UNREACHABLE_FOR_TOS              = 0xC
	COMMUNICATION_PROHIBITED_BY_FILTERING = 0xD
	HOST_PRECEDENCE_VIOLATION             = 0xE
	PRECEDENCE_CUTOFF_IN_EFFECT           = 0xF

	# Redirect (Type 5)
	REDIRECT_FOR_NETWORK                  = 0x0
	REDIRECT_FOR_HOST                     = 0x1
	REDIRECT_FOR_TOS_AND_NET              = 0x2
	REDIRECT_FOR_TOS_AND_HOST             = 0x3

	# Time Exceeded (Type 11)
	TTL_EQ_ZERO_DURING_TRANSIT            = 0x0
	TTL_EQ_ZERO_DURING_REASSEMBLY         = 0x1

	# parameter Problem (Type 12)
	REQUIRED_OPTION_MISSING               = 0x1
	IP_HEADER_BAD                         = 0x2

	def __str__ (self):
		return 'icmp code %d' % int.__str__(self)

	def __repr__ (self):
		return str(self)

def NamedICMPCode (name):
	icmp = name.lower()
	if icmp == 'communication-prohibited-by-filtering': return ICMPCode.COMMUNICATION_PROHIBITED_BY_FILTERING
	if icmp == 'destination-host-prohibited':           return ICMPCode.DESTINATION_HOST_PROHIBITED
	if icmp == 'destination-host-unknown':              return ICMPCode.DESTINATION_HOST_UNKNOWN
	if icmp == 'destination-network-prohibited':        return ICMPCode.DESTINATION_NETWORK_PROHIBITED
	if icmp == 'destination-network-unknown':           return ICMPCode.DESTINATION_NETWORK_UNKNOWN
	if icmp == 'fragmentation-needed':                  return ICMPCode.FRAGMENTATION_NEEDED
	if icmp == 'host-precedence-violation':             return ICMPCode.HOST_PRECEDENCE_VIOLATION
	if icmp == 'host-unreachable':                      return ICMPCode.HOST_UNREACHABLE
	if icmp == 'host-unreachable-for-tos':              return ICMPCode.HOST_UNREACHABLE_FOR_TOS
	if icmp == 'ip-header-bad':                         return ICMPCode.IP_HEADER_BAD
	if icmp == 'network-unreachable':                   return ICMPCode.NETWORK_UNREACHABLE
	if icmp == 'network-unreachable-for-tos':           return ICMPCode.NETWORK_UNREACHABLE_FOR_TOS
	if icmp == 'port-unreachable':                      return ICMPCode.PORT_UNREACHABLE
	if icmp == 'precedence-cutoff-in-effect':           return ICMPCode.PRECEDENCE_CUTOFF_IN_EFFECT
	if icmp == 'protocol-unreachable':                  return ICMPCode.PROTOCOL_UNREACHABLE
	if icmp == 'redirect-for-host':                     return ICMPCode.REDIRECT_FOR_HOST
	if icmp == 'redirect-for-network':                  return ICMPCode.REDIRECT_FOR_NETWORK
	if icmp == 'redirect-for-tos-and-host':             return ICMPCode.REDIRECT_FOR_TOS_AND_HOST
	if icmp == 'redirect-for-tos-and-net':              return ICMPCode.REDIRECT_FOR_TOS_AND_NET
	if icmp == 'required-option-missing':               return ICMPCode.REQUIRED_OPTION_MISSING
	if icmp == 'source-host-isolated':                  return ICMPCode.SOURCE_HOST_ISOLATED
	if icmp == 'source-route-failed':                   return ICMPCode.SOURCE_ROUTE_FAILED
	if icmp == 'ttl-eq-zero-during-reassembly':         return ICMPCode.TTL_EQ_ZERO_DURING_REASSEMBLY
	if icmp == 'ttl-eq-zero-during-transit':            return ICMPCode.TTL_EQ_ZERO_DURING_TRANSIT
	raise ValueError('unknow icmp-code %s' % icmp)



# encoding: utf-8
"""
peer.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

import sys
import time
import traceback

from bagpipe.exabgp.message              import Failure
from bagpipe.exabgp.message.nop          import NOP
from bagpipe.exabgp.message.open         import Open,Capabilities
from bagpipe.exabgp.message.update       import Update
from bagpipe.exabgp.message.keepalive    import KeepAlive
from bagpipe.exabgp.message.notification import Notification, Notify, NotConnected
from bagpipe.exabgp.network.protocol     import Protocol
from bagpipe.exabgp.processes            import ProcessError

from bagpipe.exabgp.log import Logger,LazyFormat
logger = Logger()

# As we can not know if this is our first start or not, this flag is used to
# always make the program act like it was recovering from a failure
# If set to FALSE, no EOR and OPEN Flags set for Restart will be set in the
# OPEN Graceful Restart Capability
FORCE_GRACEFUL = True

# Present a File like interface to socket.socket

class Peer (object):
	# debug hold/keepalive timers
	debug_trace = True			# debug traceback on unexpected exception
	update_time = 3

	def __init__ (self,neighbor,supervisor):
		self.supervisor = supervisor
		self.neighbor = neighbor
		# The next restart neighbor definition
		self._neighbor = None
		self.bgp = None
		# We may have new update to transmit to our peers, so we need to check
		self._updates = False

		self._loop = None
		self.open = None

		# The peer message should be processed
		self._running = False
		# The peer should restart after a stop
		self._restart = True
		# The peer was restarted (to know what kind of open to send for graceful restart)
		self._restarted = FORCE_GRACEFUL
		self._reset_skip()

		self._asn4 = True

		# The routes we have parsed from our neighbour
		self._received_routes = []

		self._route_parsed = 0L
		self._now = time.time()
		self._next_info = self._now + self.update_time

	def _reset_skip (self):
		# We are currently not skipping connection attempts
		self._skip_time = 0
		# when we can not connect to a peer how many time (in loop) should we back-off
		self._next_skip = 0

	def _more_skip (self):
		self._skip_time = time.time() + self._next_skip
		self._next_skip = int(1+ self._next_skip*1.2)
		if self._next_skip > 60:
			self._next_skip = 60

	def me (self,message):
		return "Peer %15s ASN %-7s %s" % (self.neighbor.peer_address,self.neighbor.peer_as,message)

	def stop (self):
		self._running = False
		self._restart = False
		self._restarted = False
		self._reset_skip()

	def reload (self,routes):
		self._updates = True
		self.neighbor.set_routes(routes)
		self._reset_skip()

	def restart (self,restart_neighbor=None):
		# we want to tear down the session and re-establish it
		self._running = False
		self._restart = True
		self._restarted = True
		self._route_parsed = 0L
		self._neighbor = restart_neighbor
		self._reset_skip()

	def run (self):
		if self._loop:
			try:
				if self._skip_time > time.time():
					return None
				else:
					return self._loop.next()
			except StopIteration:
				self._loop = None
		elif self._restart:
			# If we are restarting, and the neighbor definition is different, update the neighbor
			if self._neighbor:
				self.neighbor = self._neighbor
				self._neighbor = None
			self._running = True
			self._loop = self._run()
		else:
			self.bgp.close()
			self.supervisor.unschedule(self)

	def _run (self,max_wait_open=10.0):
		try:
			if self.supervisor.processes.broken(self.neighbor.peer_address):
				# XXX: we should perhaps try to restart the process ??
				raise Failure('ExaBGP lost the helper process for this peer - peer down')

			self.bgp = Protocol(self)
			self.bgp.connect()

			self._reset_skip()

			_open = self.bgp.new_open(self._restarted,self._asn4)
			logger.message(self.me('>> %s' % _open))
			yield None

			start = time.time()
			while True:
				self.open = self.bgp.read_open(_open,self.neighbor.peer_address.ip)
				if time.time() - start > max_wait_open:
					logger.message(self.me('Waited for an OPEN for too long - killing the session'))
					raise Notify(1,1,'The client took over %s seconds to send the OPEN, closing' % str(max_wait_open))
				# OPEN or NOP
				if self.open.TYPE == NOP.TYPE:
					yield None
					continue
				# This test is already done in read_open
				#if self.open.TYPE != Open.TYPE:
				#	raise Notify(5,1,'We are expecting an OPEN message')
				logger.message(self.me('<< %s' % self.open))
				if not self.open.capabilities.announced(Capabilities.FOUR_BYTES_ASN) and _open.asn.asn4():
					self._asn4 = False
					raise Notify(2,0,'peer does not speak ASN4 - restarting in compatibility mode')
				if _open.capabilities.announced(Capabilities.MULTISESSION_BGP):
					if not self.open.capabilities.announced(Capabilities.MULTISESSION_BGP):
						raise Notify(2,7,'peer does not support MULTISESSION')
					local_sessionid = set(_open.capabilities[Capabilities.MULTISESSION_BGP])
					remote_sessionid = self.open.capabilities[Capabilities.MULTISESSION_BGP]
					# Empty capability is the same as MultiProtocol (which is what we send)
					if not remote_sessionid:
						remote_sessionid.append(Capabilities.MULTIPROTOCOL_EXTENSIONS)
					remote_sessionid = set(remote_sessionid)
					# As we only send one MP per session, if the matching fails, we have nothing in common
					if local_sessionid.intersection(remote_sessionid) != local_sessionid:
						raise Notify(2,8,'peer did not reply with the sessionid we sent')
					# We can not collide due to the way we generate the configuration
				yield None
				break

			message = self.bgp.new_keepalive(force=True)
			logger.message(self.me('>> KEEPALIVE (OPENCONFIRM)'))
			yield True

			while True:
				message = self.bgp.read_keepalive()
				# KEEPALIVE or NOP
				if message.TYPE == KeepAlive.TYPE:
					logger.message(self.me('<< KEEPALIVE (ESTABLISHED)'))
					break
				yield None

			try:
				for name in self.supervisor.processes.notify(self.neighbor.peer_address):
					self.supervisor.processes.write(name,'neighbor %s up\n' % self.neighbor.peer_address)
			except ProcessError:
				# Can not find any better error code that 6,0 !
				raise Notify(6,0,'ExaBGP Internal error, sorry.')

			count = 0
			for count in self.bgp.new_announce():
				yield True
			self._updates = self.bgp.buffered()
			if count:
				logger.message(self.me('>> %d UPDATE(s)' % count))

			eor = False
			if self.neighbor.graceful_restart and \
				self.open.capabilities.announced(Capabilities.MULTIPROTOCOL_EXTENSIONS) and \
				self.open.capabilities.announced(Capabilities.GRACEFUL_RESTART):

				families = []
				for family in self.open.capabilities[Capabilities.GRACEFUL_RESTART].families():
					if family in self.neighbor.families():
						families.append(family)
				self.bgp.new_eors(families)
				if families:
					eor = True
					logger.message(self.me('>> EOR %s' % ', '.join(['%s %s' % (str(afi),str(safi)) for (afi,safi) in families])))

			if not eor:
				# If we are not sending an EOR, send a keepalive as soon as when finished
				# So the other routers knows that we have no (more) routes to send ...
				# (is that behaviour documented somewhere ??)
				c,k = self.bgp.new_keepalive(True)
				if k: logger.message(self.me('>> KEEPALIVE (no more UPDATE and no EOR)'))

			seen_update = False
			while self._running:
				self._now = time.time()
				if self._now > self._next_info:
					self._next_info = self._now + self.update_time
					display_update = True
				else:
					display_update = False

				c,k = self.bgp.new_keepalive()
				if k: logger.message(self.me('>> KEEPALIVE'))

				if display_update:
					logger.timers(self.me('Sending Timer %d second(s) left' % c))

				message = self.bgp.read_message()
				# let's read if we have keepalive before doing the timer check
				c = self.bgp.check_keepalive()

				if display_update:
					logger.timers(self.me('Receive Timer %d second(s) left' % c))

				if message.TYPE == KeepAlive.TYPE:
					logger.message(self.me('<< KEEPALIVE'))
				elif message.TYPE == Update.TYPE:
					seen_update = True
					self._received_routes.extend(message.routes)
					if message.routes:
						logger.message(self.me('<< UPDATE'))
						self._route_parsed += len(message.routes)
						if self._route_parsed:
							for route in message.routes:
								logger.routes(LazyFormat(self.me(''),str,route))
					else:
						logger.message(self.me('<< UPDATE (not parsed)'))
				elif message.TYPE not in (NOP.TYPE,):
					 logger.message(self.me('<< %d' % ord(message.TYPE)))

				if seen_update and display_update:
					logger.supervisor(self.me('processed %d routes' % self._route_parsed))
					seen_update = False

				if self._updates:
					count = 0
					for count in self.bgp.new_update():
						yield True
					logger.message(self.me('>> UPDATE (%d)' % count))
					self._updates = self.bgp.buffered()

				yield None

			if self.neighbor.graceful_restart and self.open.capabilities.announced(Capabilities.GRACEFUL_RESTART):
				logger.warning('Closing the connection without notification')
				self.bgp.close()
				return

			# User closing the connection
			raise Notify(6,3)
		except NotConnected, e:
			logger.warning('we can not connect to the peer %s' % str(e))
			self._more_skip()
			try:
				self.bgp.close()
			except Failure:
				pass
			return
		except Notify,e:
			logger.warning(self.me('Sending Notification (%d,%d) [%s] %s' % (e.code,e.subcode,str(e),e.data)))
			try:
				self.bgp.new_notification(e)
			except Failure:
				pass
			try:
				self.bgp.close()
			except Failure:
				pass
			return
		except Notification, e:
			logger.warning(self.me('Received Notification (%d,%d) from peer %s' % (e.code,e.subcode,str(e))))
			try:
				self.bgp.close()
			except Failure:
				pass
			return
		except Failure, e:
			logger.warning(self.me(str(e)),'connection')
			self._more_skip()
			try:
				self.bgp.close()
			except Failure:
				pass
			return
		except Exception, e:
			logger.warning(self.me('UNHANDLED EXCEPTION'))
			self._more_skip()
			if self.debug_trace:
				# should really go to syslog
				traceback.print_exc(file=sys.stdout)
				raise
			else:
				logger.warning(self.me(str(e)))
			if self.bgp: self.bgp.close()
			return

	def received_routes (self):
		if self._received_routes:
			while self._received_routes:
				route = self._received_routes.pop(0)
				yield str(route)
		else:
			raise StopIteration()




# encoding: utf-8
"""
protocol.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

# http://www.iana.org/assignments/protocol-numbers/
class Protocol (int):
	ICMP  = 0x01
	IGMP  = 0x02
	TCP   = 0x06
	EGP   = 0x08
	UDP   = 0x11
	RSVP  = 0x2E
	GRE   = 0x2F
	ESP   = 0x32
	AH    = 0x33
	OSPF  = 0x59
	IPIP  = 0x5E
	PIM   = 0x67
	SCTP  = 0x84

	def __str__ (self):
		if self == Protocol.ICMP:    return 'ICMP'
		if self == Protocol.IGMP:    return 'IGMP'
		if self == Protocol.TCP:     return 'TCP'
		if self == Protocol.EGP:     return 'EGP'
		if self == Protocol.UDP:     return 'UDP'
		if self == Protocol.RSVP:    return 'RSVP'
		if self == Protocol.GRE:     return 'GRE'
		if self == Protocol.ESP:     return 'ESP'
		if self == Protocol.AH:      return 'AH'
		if self == Protocol.OSPF:    return 'OSPF'
		if self == Protocol.IPIP:    return 'IPIP'
		if self == Protocol.PIM:     return 'PIM'
		if self == Protocol.SCTP:    return 'SCTP'
		return "unknown protocol %d" % int.__str__(self)

	def __repr__ (self):
		return str(self)

	def pack (self):
		return chr(self)

def NamedProtocol (protocol):
	name = protocol.upper()
	if   name == 'ICMP':  return Protocol(Protocol.ICMP)
	elif name == 'IGMP':  return Protocol(Protocol.IGMP)
	elif name == 'TCP':   return Protocol(Protocol.TCP)
	elif name == 'EGP':   return Protocol(Protocol.EGP)
	elif name == 'UDP':   return Protocol(Protocol.UDP)
	elif name == 'RSVP':  return Protocol(Protocol.RSVP)
	elif name == 'GRE':   return Protocol(Protocol.GRE)
	elif name == 'ESP':   return Protocol(Protocol.ESP)
	elif name == 'AH':    return Protocol(Protocol.AH)
	elif name == 'OSPF':  return Protocol(Protocol.OSPF)
	elif name == 'IPIP':  return Protocol(Protocol.IPIP)
	elif name == 'PIM':   return Protocol(Protocol.PIM)
	elif name == 'SCTP':  return Protocol(Protocol.SCTP)
	else: raise ValueError('unknown protocol %s' % name)



# encoding: utf-8
"""
attributes.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from struct import pack

from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

# =================================================================== MED (4)

class MED (Attribute):
	ID = AttributeID.MED
	FLAG = Flag.OPTIONAL
	MULTIPLE = False

	def __init__ (self,med):
		self.med = med

	def pack (self):
		return self._attribute(pack('!L',self.med))

	def __len__ (self):
		return 4

	def __str__ (self):
		return str(self.med)

	def __repr__ (self):
		return str(self)

	def __cmp__(self,other):
		if ( not isinstance(other,MED)
			or (self.med != other.med)
			):
			return -1
		else:
			return 0


# encoding: utf-8
"""
open.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

import os
import socket
from struct import pack

from bagpipe.exabgp.structure.address import AFI,SAFI
from bagpipe.exabgp.structure.asn  import ASN
from bagpipe.exabgp.message import Message

# =================================================================== Open

class Open (Message):
	TYPE = chr(0x01)

	def __init__ (self,version,asn,router_id,capabilities,hold_time):
		self.version = Version(version)
		self.asn = ASN(asn)
		self.hold_time = HoldTime(hold_time)
		self.router_id = RouterID(router_id)
		self.capabilities = capabilities

	def message (self):
		if self.asn.asn4():
			return self._message("%s%s%s%s%s" % (self.version.pack(),ASN(23456).pack(False),self.hold_time.pack(),self.router_id.pack(),self.capabilities.pack()))
		return self._message("%s%s%s%s%s" % (self.version.pack(),self.asn.pack(False),self.hold_time.pack(),self.router_id.pack(),self.capabilities.pack()))

	def __str__ (self):
		return "OPEN version=%d asn=%d hold_time=%s router_id=%s capabilities=[%s]" % (self.version, self.asn, self.hold_time, self.router_id,self.capabilities)

# =================================================================== Version

class Version (int):
	def pack (self):
		return chr(self)

# =================================================================== HoldTime

class HoldTime (int):
	def pack (self):
		return pack('!H',self)

	def keepalive (self):
		return int(self/3)

	def __len__ (self):
		return 2

# =================================================================== RouterID

class RouterID (object):
	def __init__ (self,ip):
		self.ip = ip
		try:
			self.raw = socket.inet_pton(socket.AF_INET,ip)
		except socket.error:
			raise ValueError('invalid IP address %s' % str(ip))

	def afi (self):
		return AFI.ipv4

	def __len__ (self):
		return 4

	def pack (self):
		return self.raw

	def __str__ (self):
		return self.ip

	def __repr__ (self):
		return str(self)

	def __eq__ (self,other):
		return self.ip == other.ip

# =================================================================== Graceful (Restart)
# RFC 4727

class Graceful (dict):
	TIME_MASK     = 0x0FFF
	FLAG_MASK     = 0xF000

	# 0x8 is binary 1000
	RESTART_STATE = 0x08
	FORWARDING_STATE = 0x80

	def __init__ (self,restart_flag,restart_time,protos):
		dict.__init__(self)
		self.restart_flag = restart_flag
		self.restart_time = restart_time & Graceful.TIME_MASK
		for afi,safi,family_flag in protos:
			self[(afi,safi)] = family_flag & Graceful.FORWARDING_STATE

	def extract (self):
		restart  = pack('!H',((self.restart_flag << 12) | (self.restart_time & Graceful.TIME_MASK)))
		families = [(afi.pack(),safi.pack(),chr(self[(afi,safi)])) for (afi,safi) in self.keys()]
		sfamilies = ''.join(["%s%s%s" % (pafi,psafi,family) for (pafi,psafi,family) in families])
		return ["%s%s" % (restart,sfamilies)]

	def __str__ (self):
		families = [(str(afi),str(safi),hex(self[(afi,safi)])) for (afi,safi) in self.keys()]
		sfamilies = ' '.join(["%s/%s=%s" % (afi,safi,family) for (afi,safi,family) in families])
		return "Graceful Restart Flags %s Time %d %s" % (hex(self.restart_flag),self.restart_time,sfamilies)

	def families (self):
		return self.keys()

# =================================================================== MultiProtocol

class MultiProtocol (list):
	def __str__ (self):
		return 'Multiprotocol ' + ' '.join(["%s %s" % (safi,ssafi) for (safi,ssafi) in [(str(afi),str(safi)) for (afi,safi) in self]])

	def extract (self):
		rs = []
		for v in self:
			rs.append(pack('!H',v[0]) + pack('!H',v[1]))
		return rs

# =================================================================== MultiSession

class MultiSession (list):
	def __str__ (self):
		return 'Multisession %s' % ' '.join([str(capa) for capa in self])

	def extract (self):
		rs = [chr(0),]
		for v in self:
			rs.append(chr(v))
		return  rs

# =================================================================== RouteRefresh

class RouteRefresh (list):
	def __str__ (self):
		return "Route Refresh (unparsed)"

	def extract (self):
		return []

class CiscoRouteRefresh (list):
	def __str__ (self):
		return "Cisco Route Refresh (unparsed)"

	def extract (self):
		return []

# =================================================================== Parameter

class ASN4 (int):
	def extract (self):
		return [pack('!L',self)]

# =================================================================== Unknown

class Unknown (object):
	def __init__ (self,value,raw=''):
		self.value = value
		self.raw = raw

	def __str__ (self):
		if self.value in Capabilities.reserved: return 'Reserved %s' % str(self.value)
		if self.value in Capabilities.unassigned: return 'Unassigned %s' % str(self.value)
		return 'unknown %s %s' % (str(self.value),str(self.raw))

	def extract (self):
		return []

# =================================================================== Parameter

class Parameter (int):
	AUTHENTIFICATION_INFORMATION = 0x01 # Depreciated
	CAPABILITIES                 = 0x02

	def __str__ (self):
		if self == 0x01: return "AUTHENTIFICATION INFORMATION"
		if self == 0x02: return "OPTIONAL"
		return 'UNKNOWN'

# =================================================================== Capabilities
# http://www.iana.org/assignments/capability-codes/

class Capabilities (dict):
	RESERVED                 = 0x00 # [RFC5492]
	MULTIPROTOCOL_EXTENSIONS = 0x01 # [RFC2858]
	ROUTE_REFRESH            = 0x02 # [RFC2918]
	OUTBOUND_ROUTE_FILTERING = 0x03 # [RFC5291]
	MULTIPLE_ROUTES          = 0x04 # [RFC3107]
	EXTENDED_NEXT_HOP        = 0x05 # [RFC5549]
	#6-63      Unassigned
	GRACEFUL_RESTART         = 0x40 # [RFC4724]
	FOUR_BYTES_ASN           = 0x41 # [RFC4893]
	# 66 Deprecated
	DYNAMIC_CAPABILITY       = 0x43 # [Chen]
	MULTISESSION_BGP_RFC     = 0x44 # [draft-ietf-idr-bgp-multisession]
	ADD_PATH                 = 0x45 # [draft-ietf-idr-add-paths]
	# 70-127    Unassigned
	CISCO_ROUTE_REFRESH      = 0x80 # I Can only find reference to this in the router logs
	# 128-255   Reserved for Private Use [RFC5492]
	MULTISESSION_BGP         = 0x83 # What Cisco really use for Multisession (yes this is a reserved range in prod !)

	EXTENDED_MESSAGE         = -1 # No yet defined by draft http://tools.ietf.org/html/draft-ietf-idr-extended-messages-02.txt

	unassigned = range(70,128)
	reserved = range(128,256)

	def announced (self,capability):
		return self.has_key(capability)

	# XXX: Should we not call the __str__ function of all the created capability classes ?
	def __str__ (self):
		r = []
		for key in self.keys():
			if key == self.MULTIPROTOCOL_EXTENSIONS:
				r += [str(self[key])]
			elif key == self.ROUTE_REFRESH:
				r += ['Route Refresh']
			elif key == self.CISCO_ROUTE_REFRESH:
				r += ['Cisco Route Refresh']
			elif key == self.GRACEFUL_RESTART:
				r += ['Graceful Restart']
			elif key == self.FOUR_BYTES_ASN:
				r += ['4Bytes AS %d' % self[key]]
			elif key == self.MULTISESSION_BGP:
				r += [str(self[key])]
			elif key == self.MULTISESSION_BGP_RFC:
				r += ['Multi Session']
			elif key in self.reserved:
				r += ['private use capability %d' % key]
			elif key in self.unassigned:
				r += ['unassigned capability %d' % key]
			else:
				r += ['unhandled capability %d' % key]
		return ', '.join(r)

	def default (self,neighbor,restarted):
		graceful = neighbor.graceful_restart

		if neighbor.multisession or os.environ.get('MINIMAL_MP','0') in ['','1','yes','Yes','YES']:
			families = neighbor.families()
		else:
			families = []
			families.append((AFI(AFI.ipv4),SAFI(SAFI.unicast)))
			families.append((AFI(AFI.ipv6),SAFI(SAFI.unicast)))
			families.append((AFI(AFI.ipv4),SAFI(SAFI.flow_ipv4)))

		mp = MultiProtocol()
		mp.extend(families)
		self[Capabilities.MULTIPROTOCOL_EXTENSIONS] = mp
		self[Capabilities.FOUR_BYTES_ASN] = ASN4(neighbor.local_as)

		if graceful:
			if restarted:
				self[Capabilities.GRACEFUL_RESTART] = Graceful(Graceful.RESTART_STATE,graceful,[(afi,safi,Graceful.FORWARDING_STATE) for (afi,safi) in families])
			else:
				self[Capabilities.GRACEFUL_RESTART] = Graceful(0x0,graceful,[(afi,safi,Graceful.FORWARDING_STATE) for (afi,safi) in families])

		# MUST be the last key added
		if neighbor.multisession:
			self[Capabilities.MULTISESSION_BGP] = MultiSession([Capabilities.MULTIPROTOCOL_EXTENSIONS])
		return self

	def pack (self):
		rs = []
		for k,capabilities in self.iteritems():
			for capability in capabilities.extract():
				rs.append("%s%s%s" % (chr(k),chr(len(capability)),capability))
		parameters = "".join(["%s%s%s" % (chr(2),chr(len(r)),r) for r in rs])
		return "%s%s" % (chr(len(parameters)),parameters)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os.path

import sys

from logging import Logger
import logging.config

# import logging_tree

import traceback

from daemon import runner
import signal

from ConfigParser import SafeConfigParser, NoSectionError
from optparse import OptionParser

from bagpipe.bgp.common import utils
from bagpipe.bgp.common.looking_glass import LookingGlass, \
    LookingGlassLogHandler

from bagpipe.bgp.engine.bgp_manager import Manager

from bagpipe.bgp.rest_api import RESTAPI

from bagpipe.bgp.vpn import VPNManager


def findDataplaneDrivers(dpConfigs, bgpConfig, isCleaningUp=False):
    drivers = dict()
    for vpnType in dpConfigs.iterkeys():
        dpConfig = dpConfigs[vpnType]

        if 'dataplane_driver' not in dpConfig:
            logging.error(
                "no dataplane_driver set for %s (%s)", vpnType, dpConfig)

        driverName = dpConfig["dataplane_driver"]
        logging.debug(
            "Creating dataplane driver for %s, with %s", vpnType, driverName)

        # FIXME: this is a hack, dataplane drivers should have a better way to
        #  access any item in the BGP dataplaneConfig
        if 'dataplane_local_address' not in dpConfig:
            dpConfig['dataplane_local_address'] = bgpConfig['local_address']

        for tentativeClassName in (driverName,
                                   'bagpipe.%s' % driverName,
                                   'bagpipe.bgp.%s' % driverName,
                                   'bagpipe.bgp.vpn.%s.%s' % (
                                       vpnType, driverName),
                                   ):
            try:
                if '.' not in tentativeClassName:
                    logging.debug(
                        "Not trying to import '%s'", tentativeClassName)
                    continue

                driverClass = utils.import_class(tentativeClassName)
                try:
                    logging.info("Found driver for %s, initiating...", vpnType)
                    # skip the init step if called for cleanup
                    driver = driverClass(dpConfig, not isCleaningUp)
                    drivers[vpnType] = driver
                    logging.info(
                        "Successfully initiated dataplane driver for %s with"
                        " %s", vpnType, tentativeClassName)
                except ImportError as e:
                    logging.debug(
                        "Could not initiate dataplane driver for %s with"
                        " %s: %s", vpnType, tentativeClassName, e)
                except Exception as e:
                    logging.error(
                        "Found class, but error while instantiating dataplane"
                        " driver for %s with %s: %s", vpnType,
                        tentativeClassName, e)
                    logging.error(traceback.format_exc())
                    break
                break
            except SyntaxError as e:
                logging.error(
                    "Found class, but syntax error while instantiating "
                    "dataplane driver for %s with %s: %s", vpnType,
                    tentativeClassName, e)
                break
            except Exception as e:
                logging.debug(
                    "Could not initiate dataplane driver for %s with %s (%s)",
                    vpnType, tentativeClassName, e)
    return drivers


class BgpDaemon(LookingGlass):

    def __init__(self, catchAllLGLogHandler, **kwargs):
        self.stdin_path = '/dev/null'
        self.stdout_path = '/dev/null'
        self.stderr_path = '/dev/null'
        self.pidfile_path = '/var/run/bagpipe-bgp/bagpipe-bgp.pid'
        self.pidfile_timeout = 5

        logging.info("BGP manager configuration : %s", kwargs["bgpConfig"])
        self.bgpConfig = kwargs["bgpConfig"]

        logging.info("BGP dataplane dataplaneDriver configuration : %s",
                     kwargs["dataplaneConfig"])
        self.dataplaneConfig = kwargs["dataplaneConfig"]

        logging.info("BGP API configuration : %s", kwargs["apiConfig"])
        self.apiConfig = kwargs["apiConfig"]

        self.catchAllLGLogHandler = catchAllLGLogHandler

    def run(self):
        logging.info("Starting BGP component...")

        logging.debug("Creating dataplane drivers")
        drivers = findDataplaneDrivers(self.dataplaneConfig, self.bgpConfig)

        for vpnType in self.dataplaneConfig.iterkeys():
            if vpnType not in drivers:
                logging.error(
                    "Could not initiate any dataplane driver for %s", vpnType)
                return

        logging.debug("Creating BGP manager")
        self.bgpManager = Manager(self.bgpConfig)

        logging.debug("Creating VPN manager")
        self.vpnManager = VPNManager(self.bgpManager, drivers)

        # BGP component REST API
        logging.debug("Creating REST API")
        bgpapi = RESTAPI(
            self.apiConfig, self, self.vpnManager, self.catchAllLGLogHandler)
        bgpapi.run()

    def stop(self, signum, frame):
        logging.info("Received signal %(signum)r, stopping...", vars())
        self.vpnManager.stop()
        self.bgpManager.stop()
        # would need to stop main thread ?
        logging.info("All threads now stopped...")
        exception = SystemExit("Terminated on signal %(signum)r" % vars())
        raise exception

    def getLookingGlassLocalInfo(self, pathPrefix):
        return {
            "dataplane": self.dataplaneConfig,
            "bgp": self.bgpConfig
        }


def _loadConfig(configFile):
    parser = SafeConfigParser()

    if (len(parser.read(configFile)) == 0):
        logging.error("Configuration file not found (%s)", configFile)
        exit()

    bgpConfig = parser.items("BGP")

    dataplaneConfig = dict()
    for vpnType in ['ipvpn', 'evpn']:
        try:
            dataplaneConfig[vpnType] = dict(
                parser.items("DATAPLANE_DRIVER_%s" % vpnType.upper()))
        except NoSectionError:
            if vpnType == "ipvpn":  # backward compat for ipvpn
                dataplaneConfig['ipvpn'] = dict(
                    parser.items("DATAPLANE_DRIVER"))
                logging.warning("Config file is obsolete, should have a "
                                "DATAPLANE_DRIVER_IPVPN section instead of"
                                " DATAPLANE_DRIVER")
            else:
                logging.error(
                    "Config file should have a DATAPLANE_DRIVER_EVPN section")

    apiConfig = parser.items("API")
    # TODO: add a default API config

    config = {"bgpConfig": dict(bgpConfig),
              "dataplaneConfig": dataplaneConfig,
              "apiConfig": dict(apiConfig)
              }

    return config


def daemon_main():
    usage = "usage: %prog [options] (see --help)"
    parser = OptionParser(usage)

    parser.add_option("--config-file", dest="configFile",
                      help="Set BGP component configuration file path",
                      default="/etc/bagpipe-bgp/bgp.conf")
    parser.add_option("--log-file", dest="logFile",
                      help="Set logging configuration file path",
                      default="/etc/bagpipe-bgp/log.conf")
    parser.add_option("--no-daemon", dest="daemon", action="store_false",
                      help="Do not daemonize", default=True)
    (options, _) = parser.parse_args()

    action = sys.argv[1]
    assert(action == "start" or action == "stop")

    if not os.path.isfile(options.logFile):
        logging.basicConfig()
        print "no logging configuration file at %s" % options.logFile
        logging.warning("no logging configuration file at %s", options.logFile)
    else:
        logging.config.fileConfig(
            options.logFile, disable_existing_loggers=False)

    if action == "start":
        logging.root.name = "Main"
        logging.info("Starting...")
    else:  # stop
        logging.root.name = "Stopper"
        logging.info("Signal daemon to stop")

    catchAllLogHandler = LookingGlassLogHandler()

    # we inject this catch all log handler in all configured loggers
    for (loggerName, logger) in Logger.manager.loggerDict.iteritems():
        if isinstance(logger, Logger):
            if (not logger.propagate and logger.parent is not None):
                logging.debug("Adding looking glass log handler to logger: %s",
                              loggerName)
                logger.addHandler(catchAllLogHandler)
    logging.root.addHandler(catchAllLogHandler)

    # logging_tree.printout()

    config = _loadConfig(options.configFile)

    bgpDaemon = BgpDaemon(catchAllLogHandler, **config)

    try:
        if options.daemon:
            daemon_runner = runner.DaemonRunner(bgpDaemon)
            # This ensures that the logger file handler does not get closed
            # during daemonization
            daemon_runner.daemon_context.files_preserve = [
                logging.getLogger().handlers[0].stream]
            daemon_runner.daemon_context.signal_map = {
                signal.SIGTERM: bgpDaemon.stop
            }
            daemon_runner.do_action()
        else:
            signal.signal(signal.SIGTERM, bgpDaemon.stop)
            signal.signal(signal.SIGINT, bgpDaemon.stop)
            bgpDaemon.run()
    except Exception as e:
        logging.exception("Error while starting BGP daemon: %s", e)

    logging.info("BGP component main thread stopped.")


def cleanup_main():
    usage = "usage: %prog [options] (see --help)"
    parser = OptionParser(usage)

    parser.add_option("--config-file", dest="configFile",
                      help="Set BGP component configuration file path",
                      default="/etc/bagpipe-bgp/bgp.conf")
    parser.add_option("--log-file", dest="logFile",
                      help="Set logging configuration file path",
                      default="/etc/bagpipe-bgp/log.conf")
    (options, _) = parser.parse_args()

    if not os.path.isfile(options.logFile):
        print "no logging configuration file at %s" % options.logFile
        logging.basicConfig()
    else:
        logging.config.fileConfig(
            options.logFile, disable_existing_loggers=False)
        logging.root.name = "[BgpDataplaneCleaner]"

    logging.info("Cleaning BGP component dataplanes...")
    config = _loadConfig(options.configFile)

    drivers = findDataplaneDrivers(
        config["dataplaneConfig"], config["bgpConfig"], isCleaningUp=True)
    for (vpnType, dataplaneDriver) in drivers.iteritems():
        logging.info("Cleaning BGP component dataplane for %s...", vpnType)
        dataplaneDriver.resetState()

    logging.info("BGP component dataplanes have been cleaned up.")

if __name__ == '__main__':
    daemon_main()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

from threading import Thread, Event, Timer

import time
from time import sleep

import logging
import traceback

from bagpipe.bgp.engine.worker import Worker
from bagpipe.bgp.engine import RouteEvent

from bagpipe.bgp.common.looking_glass import LookingGlassLocalLogger

Init = "InitEvent"
ReInit = "ReInit"
SendKeepAlive = "Send KeepAlive"
KeepAliveReceived = "KeepAlive-received"

DEFAULT_HOLDTIME = 180


class FSM(object):

    '''
    Represents the state of the BGP Finite State Machine
    '''

    Idle = "Idle"
    Connect = 'Connect'
    OpenSent = 'OpenSent'
    OpenConfirm = 'OpenConfirm'
    Active = 'Active'
    Established = 'Established'

    def __init__(self, worker):
        self.worker = worker
        self._state = FSM.Idle
        self._prevState = None
        self.allStates = [FSM.Idle, FSM.Connect, FSM.OpenSent,
                          FSM.OpenConfirm, FSM.Active, FSM.Established]
        self.lastTransitionTime = time.time()

    @property
    def state(self):
        return self._state

    @property
    def previousState(self):
        return self._prevState

    @state.setter
    def state(self, state):
        if state in self.allStates:
            self._prevState = self._state
            self._state = state
            self.lastTransitionTime = time.time()
            self.worker.log.info(
                "%s BGP FSM transitioned from '%s' to '%s' state" %
                (self.worker, self._prevState, self._state))
        else:
            raise Exception("no such state (%s)" % repr(state))

    def __repr__(self):
        return self._state


class StoppedException(Exception):
    pass


class InitiateConnectionException(Exception):
    pass


class OpenWaitTimeout(Exception):
    pass


class BGPPeerWorker(Worker, Thread, LookingGlassLocalLogger):
    __metaclass__ = ABCMeta

    '''
    Partially abstract class for a Worker implementing the BGP protocol.
    '''

    def __init__(self, bgpManager, name, peerAddress):
        # call super
        Thread.__init__(self)
        self.setDaemon(True)
        self.name = "BGP-%s" % peerAddress
        Worker.__init__(self, bgpManager, self.name)

        self.bgpManager = bgpManager
        self.peerAddress = peerAddress

        # its up to subclasses to call setHoldTime again to set holdtime based
        # on value advertized by peer
        self._setHoldTime(DEFAULT_HOLDTIME)

        # used to stop receiveThread
        self._stopLoops = Event()
        # used to track that we've been told to stop:
        self.shouldStop = False

        self.sendKATimer = None
        self.KAReceptionTimer = None

        LookingGlassLocalLogger.__init__(
            self, self.peerAddress.replace(".", "-"))

        self.fsm = FSM(self)

        self.log.debug("Init %s", self.name)
        self.enqueue(Init)

    def stop(self):
        Worker.stop(self)
        self._stopLoops.set()
        self.shouldStop = True

    def _setHoldTime(self, holdtime):
        '''
        holdtime in seconds
        keepAlive expected, or sent, every holdtime/3 second
        '''
        assert(holdtime > 30)
        self.katPeriod = int(holdtime / 3.0)
        self.katExpiryTime = self.katPeriod * 3

    # called by _eventQueueProcessorLoop
    def _onEvent(self, event):

        if event == Init:
            self._initiate()

        elif event == ReInit:
            self._reinitiate()

        elif isinstance(event, RouteEvent):
            if (self.fsm.state == FSM.Established):
                self._send(self._updateForRouteEvent(event))
            else:
                # FIXME: this is possibly not correct yet: why did we received
                # this event  ? what do we do with it ?
                raise Exception(
                    "cannot process routeEvent in '%s' state" % self.fsm.state)

        elif event == SendKeepAlive:
            self._send(self._keepAliveMessageData())

        elif event == KeepAliveReceived:
            self.onKeepAliveReceived()

        else:
            self.log.warning("event not processed: %s", event)

    def _stopped(self):
        self.fsm.state = FSM.Idle

    def _initiate(self):
        self.log.info("Initiating")
        self._initiateConnectionAndThreads()

    def _initiateConnectionAndThreads(self):
        self._resetLocalLGLogs()
        # initiate connection

        self.fsm.state = FSM.Connect

        try:
            self._initiateConnection()
        except (InitiateConnectionException, OpenWaitTimeout) as e:
            self.log.warning(
                "%s while initiating connection: %s", e.__class__.__name__, e)
            self.enqueue(ReInit)
            return
            # FIXME: we should transition to Active or Idle state depending how
            # many times we already tried
        except StoppedException:
            self.log.info("Stopped during connection init")
            return
        except Exception as e:
            self.log.warning("Exception while initiating connection: %s", e)
            if self.log.isEnabledFor(logging.DEBUG):
                self.log.debug("%s", traceback.format_exc())
            self.enqueue(ReInit)
            # FIXME: we should transition to Active or Idle state depending how
            # many times we already tried
            return

        self._stopLoops.clear()

        self.initSendKeepAliveTimer()
        self.initKeepAliveReceptionTimer()

        # spawns a receive thread
        self.receiveThread = Thread(target=self._receiveLoop,
                                    name="%s:receiveLoop" % self.name)
        self.receiveThread.start()

    def _toEstablished(self):
        self.fsm.state = FSM.Established

    def _toIdle(self):
        pass

    def _reinitiate(self):
        self.log.info("Re-initiating")

        self.fsm.state = FSM.Idle

        self._stopLoops.set()

        if self.sendKATimer:
            self.sendKATimer.cancel()
        if self.KAReceptionTimer:
            self.KAReceptionTimer.cancel()

        self.bgpManager.cleanup(self)

        self._toIdle()

        # TODO(tmmorin): read BGP specs to get the retries timers right
        # TODO(tmmorin): replace with a timer that injects an event, so that we
        # avoid sleeping which make us miss any stop event
        sleep(10)

        self._initiateConnectionAndThreads()

    def isEstablished(self):
        return (self.fsm.state == FSM.Established)

    def _receiveLoop(self):
        self.log.info("Start receive loop")
        while not self._stopLoops.isSet():
            try:
                loopResult = self._receiveLoopFun()
                if loopResult == 0:
                    self.log.info(
                        "receiveLoopFun returned 0, aborting receiveLoop")
                    break
                elif loopResult == 2:
                    self.log.warning("receiveLoopFun returned 2 (error), "
                                     "aborting receiveLoop and reinitializing")
                    # FIXME: use (Worker.)enqueueHighPriority so that
                    # ReInit is treated before other events
                    self.enqueue(ReInit)
                    break
                else:
                    # everything went fine
                    pass
            except Exception as e:
                self.log.error("Error: %s (=> aborting receiveLoop and "
                               "reinitializing)", e)
                if self.log.isEnabledFor(logging.WARNING):
                    self.log.warning("%s", traceback.format_exc())
                # FIXME: use (Worker.)enqueueHighPriority so that
                # ReInit is treated before other events
                self.enqueue(ReInit)
                break

        self.log.info("End receive loop")

    # Sending keep-alive's #####

    def initSendKeepAliveTimer(self):
        self.log.debug("Init sendKA timer (%ds)", self.katPeriod)
        self.sendKATimer = Timer(
            self.katPeriod, BGPPeerWorker.sendKeepAliveTrigger, [self])
        self.sendKATimer.name = "%s:sendKATimer" % self.name
        self.sendKATimer.start()

    def sendKeepAliveTrigger(self):
        self.log.debug("Trigger send KeepAlive")
        self.enqueue(SendKeepAlive)
        self.initSendKeepAliveTimer()

    # Receiving keep-alive's #####

    def initKeepAliveReceptionTimer(self):
        self.log.debug(
            "Init Keepalive reception timer (%ds)", self.katExpiryTime)
        self.KAReceptionTimer = Timer(
            self.katExpiryTime, BGPPeerWorker.onKeepAliveExpired, [self])
        self.KAReceptionTimer.start()

    def onKeepAliveReceived(self):
        self.log.debug("Keepalive received")
        self.KAReceptionTimer.cancel()
        self.initKeepAliveReceptionTimer()

    def onKeepAliveExpired(self):
        self.log.error("KeepAlive expired, re-init")
        self.fsm.state = FSM.Idle
        self.enqueue(ReInit)

    # Abstract methods

    @abstractmethod
    def _initiateConnection(self):
        '''
        Abstract method.
        The implementation will initiated the connection to the BGP peer, do
        the initial BGP handshake (send Open, receive Open, send first
        KeepAlive, receive first KeepAlive) and track the intermediate
        FSM states (OpenSent, OpenConfirm).
        '''
        pass

    @abstractmethod
    def _receiveLoopFun(self):
        '''
        Return codes:
        - 0: we decided to stop based on stopLoop being set
        - 1: everything went fine, pursue
        - 2: there was an error
        '''
        pass

    @abstractmethod
    def _keepAliveMessageData(self):
        pass

    @abstractmethod
    def _send(self, data):
        pass

    @abstractmethod
    def _updateForRouteEvent(self, event):
        pass

    # Looking glass hooks ###

    def getLookingGlassLocalInfo(self, pathPrefix):
        return {
            "protocol": {
                "state": self.fsm.state,
                "previous_state": "(%s)" % self.fsm.previousState,
                "hold_time": self.katExpiryTime,
                "last_transition_time": time.strftime(
                    '%Y-%m-%d %H:%M:%S',
                    time.localtime(self.fsm.lastTransitionTime))
            }
        }



# encoding: utf-8
"""
utils.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

#import sys
#import time

import StringIO
import traceback

def hexa (value):
	return "%s" % [(hex(ord(_))) for _ in value]

def hexdump (value):
	print hexa(value)

def trace ():
	buff = StringIO.StringIO()
	traceback.print_exc(file=buff)
	r = buff.getvalue()
	buff.close()
	return r



"""
Copyright (c) 2014, Orange
All rights reserved.

File released under the BSD 3-Clause license.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions 
are met:

1. Redistributions of source code must retain the above copyright 
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in 
   the documentation and/or other materials provided with the 
   distribution.

3. Neither the name of the copyright holder nor the names of its 
   contributors may be used to endorse or promote products derived 
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
"""

from struct import pack,unpack

from bagpipe.exabgp.structure.asn import ASN
from bagpipe.exabgp.structure.address    import AFI,SAFI
from bagpipe.exabgp.message.update.attribute.communities import RouteTarget

class RouteTargetConstraint(object):
    # TODO: no support yet for RTC variable length with prefixing
    
    def __init__(self,afi,safi,origin_as,route_target):
        self.afi = AFI(afi)
        self.safi = SAFI(safi)
        self.origin_as = origin_as
        self.route_target = route_target
        
    def __len__(self):
        if self.route_target is None:
            return 1
        else:
            return (5 + len(self.route_target))
    
    def __str__ (self):
        if self.route_target is None:
            return "RTC Wildcard"
        else:
            return "RTC<%s>:%s" % ( self.origin_as, self.route_target )
    
    def __repr__(self):
        return self.__str__() 
    
    def __cmp__(self,other):
        if (isinstance(other,RouteTargetConstraint) and
            self.origin_as == other.origin_as and
            self.route_target == other.route_target):
            return 0
        else:
            return -1
        
    def __hash__(self):
        return hash(self.pack())

    @staticmethod
    def resetFlags(char):
        return chr(ord(char) & ~(0x40))

    def pack(self):
        if self.route_target ==  None:
            return pack("!B",0)
        else:
            packedRT = self.route_target.pack()
            # We reset ext com flag bits from the first byte in the packed RT
            # because in an RTC route these flags never appear.
            return pack("!BL", (len(self)-1)*8, self.origin_as) + RouteTargetConstraint.resetFlags(packedRT[0]) + packedRT[1:]
        
    @staticmethod
    def unpack(afi,safi,data):
        len_in_bits = ord(data[0]) 
        data=data[1:]
                
        if (len_in_bits==0):
            return RouteTargetConstraint(afi,safi,ASN(0),None)
        
        if (len_in_bits<4):
            raise Exception("RTC route too short to be decoded (len %d bits)" % len_in_bits)
        
        asn = ASN( unpack('!L', data[0:4] )[0] )
        data = data[4:]
        
        rt = RouteTarget.unpackFrom(data)
        return RouteTargetConstraint(afi,safi,asn,rt)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

import traceback

from Queue import Queue

from threading import Event

from bagpipe.bgp.engine import RouteEntry, RouteEvent, \
    Subscription, Unsubscription
from bagpipe.bgp.common.looking_glass import LookingGlass, LGMap

log = logging.getLogger(__name__)


class Worker(LookingGlass):

    """This is the base class for objects that interact with the route table
    manager to produce and consume events related to BGP routes.

    These objects will:
    * use _subscribe(...) and _unsubscribe(...) to subscribe to routing events
    * will specialize _onEvent(event) to react to received events
    * use _pushEvent(event) to publish routing events

    """

    stopEvent = object()

    def __init__(self, bgpManager, workerName):
        self.bgpManager = bgpManager
        self._queue = Queue()
        self._pleaseStop = Event()

        log.debug("Setting worker name to %s", workerName)
        self.name = workerName
        assert(self.name is not None)

        log.debug("Instantiated %s worker", self.name)

    def stop(self):
        """
        Stop this worker.

        Set the _pleaseStop internal event to stop the event processor loop
        and indicate to the route table manager that this worker is stopped.
        Then call _stopped() to let a subclass implement any further work.
        """
        self._pleaseStop.set()
        self._queue.put(Worker.stopEvent)
        self.bgpManager.cleanup(self)
        self._stopped()

    def _stopped(self):
        """
        Hook for subclasses to react when Worker is stopped (NoOp in base
        Worker class)
        """

    def _eventQueueProcessorLoop(self):
        """
        Main loop where the worker consumes events.
        """
        while not self._pleaseStop.isSet():
            # log.debug("%s worker waiting on queue",self.name )
            event = self._dequeue()

            if (event == Worker.stopEvent):
                log.debug("StopEvent, breaking queue processor loop")
                self._pleaseStop.set()
                break

            # log.debug("%s worker calling _onEvent for %s",self.name,event)
            try:
                self._onEvent(event)
            except Exception as e:
                log.error("Exception raised on subclass._onEvent: %s", e)
                log.error("%s", traceback.format_exc())

    def run(self):
        self._eventQueueProcessorLoop()

    def _onEvent(self, event):
        """
        This method is implemented by subclasses to react to routing events.
        """
        log.debug("Worker %s _onEvent: %s", self.name, event)
        raise NotImplementedError

    def _dequeue(self):
        return self._queue.get()

    def enqueue(self, event):
        # TODO(tmmorin): replace Queue by a PriorityQueue and use a higher
        # priority for ReInit event
        self._queue.put(event)

    def _subscribe(self, afi, safi, rt=None):
        subobj = Subscription(afi, safi, rt, self)
        log.info("Subscribe: %s ", subobj)
        self.bgpManager.routeEventSubUnsub(subobj)

    def _unsubscribe(self, afi, safi, rt=None):
        subobj = Unsubscription(afi, safi, rt, self)
        log.info("Unsubscribe: %s ", subobj)
        self.bgpManager.routeEventSubUnsub(subobj)

    def getWorkerSubscriptions(self):
        return self.bgpManager.routeTableManager.getWorkerSubscriptions(self)

    def getWorkerRouteEntries(self):
        return self.bgpManager.routeTableManager.getWorkerRouteEntries(self)

    def _pushEvent(self, routeEvent):
        assert(isinstance(routeEvent, RouteEvent))
        log.debug("Pushing route event to BGPManager")
        if routeEvent.source is None:
            routeEvent.source = self
        self.bgpManager._pushEvent(routeEvent)

    def _newRouteEntry(self, afi, safi, rts, nlri, attributes):
        return RouteEntry(afi, safi, rts, nlri, attributes, self)

    def __repr__(self):
        return "Worker %s" % (self.name)

    # Looking glass ###

    def getLookingGlassLocalInfo(self, pathPrefix):
        return {
            "name": self.name,
            "internals": {
                "event queue length": self._queue.qsize(),
                "subscriptions":
                    [repr(sub) for sub in self.getWorkerSubscriptions()],
            }
        }

    def getLGMap(self):
        return {
            "routes": (LGMap.SUBTREE, self.getLGRoutes)
        }

    def getLGRoutes(self, pathPrefix):
        return [route.getLookingGlassInfo(pathPrefix) for route in
                self.getWorkerRouteEntries()]



# encoding: utf-8
"""
attributes.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

# =================================================================== NextHop (3)

class NextHop (Attribute):
	ID = AttributeID.NEXT_HOP
	FLAG = Flag.TRANSITIVE
	MULTIPLE = False

	# Take an IP as value
	def __init__ (self,next_hop):
		self.next_hop = next_hop

	def pack (self):
		return self._attribute(self.next_hop.pack())

	def __len__ (self):
		return len(self.next_hop.pack())

	def __str__ (self):
		return str(self.next_hop)

	def __repr__ (self):
		return str(self)

	def __cmp__(self,other):
		if ( not isinstance(other,NextHop) or
			 (self.next_hop.pack() != other.next_hop.pack()) 
			):
			return -1
		else:
			return 0